""" MyHist object

Created by Philippe COTTE, cottephi@gmail.com, https://gitlab.com/cottephi

Requieres toolbox module (part of the same gitlab project)

Uses numpy.histogram objects to compute bin contents and edges from a list of
data, with the possibility to specify weights for each entry. Weights will
be considered as seperate entries.
Has numerous methods allowing to update the bin contents later on, based on
the TH1 histogram of Root (CERN software).
Has a build-in plot function printing all relevent information of the histo
(mean, rms, overflow, underflow, counts, integral, fits if any...)
Has a built-in method to fit the histogram with one or two gaussians, or with
a user-specified method (uses eval to read a string.  Will be improved to
accept an full method as argument).
Can be saved to hdf5 file that can be re-read later to recreate the
histogram object in a different script.
Can read data from simple text file, supporting a lot of options to choose
which columns to read, add or substract.

  Class attributes:
    Created in all creation methods: 
      __id -- integer unique to each instance of the class.
      __name -- name of the object.
      __data -- raw data used to compute the histogram.
      __bins -- number of bins. Can be 'auto'
      __arange  -- x-axis range.
      __density -- bool saying whether the histogram has been normalised.
      __weights -- weights to apply to the raw data.
      __ndecimals -- precisions of printed numerical attributes.
      __fixed_bin_width -- bin width specified by the user.
      __up_to_date -- says whether the statistical attritbutes are up to date.
      __no_name_numbering -- by default, if another MyHist exists with the same
        name,name will be appened with _n, n being the number of occurence of
        this name. This option disables that. If no name is specified, the
        histogram will still be named according to the number of MyHist instances.
      
    Created in reinitialise, read and copy methods:
      Histogram attributes:
        __bin_width -- actual bin width, computed from bin_edges.  Can be
          slightly different from fixed_bin_width because of roundings.
        __bin_edges -- the x position of the left side of each bin, plus
          the x position of the right side of the last bin
          (length is len(bin_middles) + 1)).
        __bin_middles -- the x position of the middle of each bin.
        __bin_contents -- counts in each bin.
        __underflow -- number of entries below range.
        __overflow -- number of entries above range.
      Statistical attributes:
        __mean -- mean of bin content*bin_middle.
        __rms -- standard deviation.
        __integral -- sum of content*bin_width.
        __count -- total number of entries.
        __statbox -- string used to display the statistical attributes.
      Fit attributes:
        __fits -- list of all fits that were done (strings containing
          either the name of the method used or the methmatical expression).
        __xyfits -- list of list of pairs (x,y) of the fitted function.
        __fitted_params -- list of list of parameters for each fit.
        __fitboxes -- list of strings used to dispay the fit results in termes
          of paramters value.
        __chisquares -- list of chisquares of each fit.
  
  Public Methods:
    All getters return a copy of the class attribute unless specified
    otherwise.
    Not written yet
"""

import warnings
import numpy as np
import os
import warnings
import math
import getopt
import collections
import numbers
import cycler
import h5py
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.backends.backend_pdf as pltpdf
import matplotlib.gridspec as gridspec
from matplotlib.ticker import FormatStrFormatter
from matplotlib.font_manager import FontProperties
from decimal import Decimal
from copy import deepcopy

import toolbox as tb

mpl.rcParams['text.usetex'] = True
mpl.rcParams['text.latex.preamble'] = [r'\usepackage{amsmath}\boldmath'] #for \text

class MyHist:
  # Color blind-friendly color cycle
  CB_color_cycle = cycler.cycler(color=["#377eb8", "#ff7f00", "#4daf4a",
                                        "#f781bf", "#a65628", "#984ea3",
                                        "#999999", "#e41a1c", "#dede00"])
  mpl.rcParams["axes.prop_cycle"] = CB_color_cycle

  # Count the number of class instances created and store their names
  names = []
  skip = ""

  @classmethod
  def get_instances(cls):
    return MyHist.names[:]

  @classmethod
  def clear(cls):
    MyHist.names = []

  @classmethod
  def set_skip(cls, skip):
    MyHist.skip = skip


# Creation methods
  def __init__(self, **kwargs):
    """ Class creator.
    
    Keyword arguments:
    filename -- a string that is passed to the read() method to create the
      histogram from a hdf5 file containing a previously saved histogram. Only
      compatible with 'name'. (default '')
    name -- a string, name of this instance of the class.  Will be appended
      with _integer if the name was already taken by another instance of the
      class, unless no_name_numbering was specified.
      (default is the current number of instances of the class)
    no_name_numbering -- a boolean, prevents appending of number after the name.
      (default is False)
    data -- list of data to use to plot the histogram. (default [])
    bins -- the number of bins to use. Not compatible with 'binwidth'. Default
      is 'auto', which uses numpy.histogram automatic bin number calculation.
    arange -- Range of the histogram's x-axis.  It is a 2-elements list.
      Supports the use of string that will be converted to decimals,
      which is recommanded to avoid floating point imprecisions. Default will
      be the min and max of 'data' at histogram computation.
    density -- boolean to tell if the histogram is to be normalised.
      (default False)
    weights -- will apply a weight to each entry in 'data'.  Must have the
      same length than 'data'.  Can not be used along with automatic binning.
      (default [])
    ndecimals -- integer, determines the precision for stat and fit boxes.
      (default 2)
    binwidth -- will force the bin width. Not compatible with 'bins'.
      (default -1)
    
    """
    
    expected = ["filename", "name", "data", "bins", "arange", "ndecimals",
                "binwidth", "weights", "density", "no_name_numbering"]
    tb.check_kwargs(expected, kwargs)
    
    filename = kwargs.get("filename", "")
    self.__name = kwargs.get("name", "")
    self.__no_name_numbering = kwargs.get("no_name_numbering", False)
    if not self.__no_name_numbering:
      i = 2
      tmp_name = self.__name
      while tmp_name in MyHist.names:
        tmp_name = self.__name
        tmp_name += "-"+str(i)
        i += 1
      self.__name = tmp_name
    # If specified, will read the histogram from input file
    if filename != "":
      self.read(filename)
      return

    self.__id = len(MyHist.names)
    if self.__name == "":
      self.__name = str(self.__id)
    self.__data = list(kwargs.get("data", []))
    self.__bins = kwargs.get("bins", "auto")
    if (isinstance(self.__bins, str) and
        self.__bins != "auto"):
      raise ValueError("bins argument is invalide.")
    if (not isinstance(self.__bins, int) and
        not isinstance(self.__bins, str)):
      raise ValueError("bins must be 'auto' or int.")
    elif self.__bins == "auto":
      if int(np.version.version.split(".")[0]) < 1:
        raise ImportError("Need numpy version greater or "\
                          "equal to 1.8 to use auto bins.")
      elif int(np.version.version.split(".")[1]) < 8:
        raise ImportError("Need numpy version greater or "\
                          "equal to 1.8 to use auto bins.")
    self.__arange = kwargs.get("arange", [])
    if len(self.__arange) == 2:
      try:
        self.__arange[0] = Decimal(self.__arange[0])
        self.__arange[1] = Decimal(self.__arange[1])
      except:
        raise ValueError("Wrong arange format. Must be convertible to "\
                         "Decimal.")
    elif len(self.__arange) != 0:
      raise ValueError("Wrong arange format. Must be a list"\
                      " of two elements or an empty list.")
    self.__density = kwargs.get("density", False)
    self.__weights = kwargs.get("weights", [])
    self.__ndecimals = kwargs.get("ndecimals", 2)
    self.__fixed_bin_width = kwargs.get("binwidth", -1)
    if isinstance(self.__fixed_bin_width, str):
      try:
        self.__fixed_bin_width = Decimal(self.__fixed_bin_width)
      except:
        raise ValueError("Can not convert", self.__fixed_bin_width,
                         "to decimals.")
    if self.__bins != "auto" and self.__fixed_bin_width != -1:
      raise tb.ArgError("Can not specify number of bins "\
                              "and bin width at the same time.")

    self.reinitialise(True)  # Will initiate the other attributes
    self.__up_to_date = False
    self._update()
    MyHist.names.append(self.__name)

  def copy(self):
    """ Copy method
    
    Uses the save and read methods. Need to be able to edit /tmp
    """
    
    self.save(filename="/tmp/MyHist.mh")
    new_h = MyHist(filename="/tmp/MyHist.mh")
    os.system("rm /tmp/MyHist.mh")
    return new_h


  def read(self, filename, keep_name = False):
    """ Read the histogram from filename (string of the path to the file) """
    
    if not os.path.isfile(filename):
      raise OSError("File", filename, "not found.")
    with h5py.File(filename, "r") as ifile:
      for dsetname in ifile:
        dset = ifile[dsetname]
        if dsetname == "name":
          if not keep_name:
            self.__name = dset[0].decode("UTF-8")
        elif dsetname == "id_ndecimals":
          self.__id = int(dset[0])
          self.__ndecimals = int(dset[1])
        elif dsetname == "bins_str":
          self.__bins = dset[0].decode("UTF-8")
        elif dsetname == "bins_int":
          self.__bins = int(dset[0])
        elif dsetname == "bin_width":
          self.__bin_width = float(dset[0])
        elif dsetname == "fixed_bin_width_str":
          self.__fixed_bin_width = dset[0].decode("UTF-8")
        elif dsetname == "fixed_bin_width_f":
          self.__fixed_bin_width = float(dset[0])
        elif dsetname == "statistics":
          self.__mean = float(dset[0])
          self.__rms = float(dset[1])
          self.__integral = float(dset[2])
          self.__underflow = float(dset[3])
          self.__overflow = float(dset[4])
          self.__count = float(dset[5])
        elif dsetname == "statbox":
          self.__statbox = dset[0].decode("UTF-8")
        elif dsetname == "bools":
          self.__no_name_numbering = bool(dset[0])
          self.__up_to_date = bool(dset[1])
          self.__density = bool(dset[2])
        elif dsetname == "data":
          self.__data = [float(d) for d in list(dset)]
        elif dsetname == "arange_str":
          self.__arange = [Decimal(a.decode("UTF-8")) for a in dset]
        elif dsetname == "arange_float":
          self.__arange = [float(d) for d in list(dset)]
        elif dsetname == "weights":
          self.__weights = [float(d) for d in list(dset)]
        elif dsetname == "bin_edges":
          self.__bin_edges = [float(d) for d in list(dset)]
        elif dsetname == "bin_middles":
          self.__bin_middles = [float(d) for d in list(dset)]
        elif dsetname == "bin_contents":
          self.__bin_contents = [float(d) for d in list(dset)]
        elif dsetname == "fits":
          self.__fits = [a.decode("UTF-8") for a in dset]
        elif dsetname == "xyfits":
          self.__xyfits = deepcopy(list(dset))
        elif dsetname == "fitted_params":
          self.__fitted_params = deepcopy(list(dset))
        elif dsetname == "fitboxes":
          self.__fitboxes = [f.decode("UTF-8") for f in dset]
        elif dsetname == "chisquares":
          self.__chisquares = [float(d) for d in list(dset)]
        else:
          raise ValueError("Unkown MyHist attribute", dsetname,".")


  def load_file(self, filename, path="", option=""):
    """ Fill data from a simple text file.
    
    Arguments are:
      filename -- name or path to the file
      path -- path to the directory of the file. Only use if filename is only
        the name of the file. (default '')
      option -- options to be used to read the file, format must be
        compatible with getopt(). (default '')
        
    By default, will use space as delimiter, and only first column as data.
    Options are:
      -f must be followed by c, l or both (c, l).  If c is specified,
         first column is ignored.  If l is specified, first line is ignored.
      -c must be followed by the column indices to use to fill data, separated
         by commas (, ).  If -f option ignored first column, the indices are 
         shifted by one (the first data column will have index 0).
         Can also use 'all' to use all columns
      -w must also be followed by the column indices to use as weights for 
         the data.  Must specify as many columns as -c.  Can not be used 
         with -c all
      -d followed by a single caracter that will become the default delimiter.
      -s followed by column indices.  All the values in the specified columns
         will be substracted to the values at the same line in the data 
         columns.  At this time, it is not possible to substract one set of 
         column to one data column and another set of columns to another data 
         column.
      -a followed by column indices.  Like -s but addition instead of
         substraction
      -m multiplicative factor. Must be either one number, or several numbers
         separated by comma. If one number is entered, all columns will be
         multiplied. Else, the numbers starting from the
         second position indicate the columns to multiply by the first number.
    """
    
    ignore_first_column = False
    ignore_first_line = False
    index_data = [0]
    index_substract = []
    index_add = []
    index_weights = []
    multiply = None
    delim = " "
    
    try:
      opts, args = getopt.getopt(option.split(), "f:c:w:d:s:a:m:", [])
    except getopt.getoptError as err:
      print(str(err))
    for opt, arg in opts:
      if opt == "-f":
        arg = arg.split(",")
        for sub_arg in arg:
          if sub_arg == "c":
            ignore_first_column = True
          elif sub_arg == "l":
            ignore_first_line = True
          else:
            raise ValueError("Argument", sub_arg, "not recognized.")
      elif opt == "-c":
        index_data = []
        arg = arg.split(",")
        for sub_arg in arg:
          if sub_arg == "all":
            index_data = "all"
            break
          else:
            try:
              index_data.append(int(sub_arg))
            except:
              raise ValueError("Argument", sub_arg, "not recognized.")
      elif opt == "-w":
        arg = arg.split(",")
        for sub_arg in arg:
          try:
            index_weights.append(int(sub_arg))
          except:
            raise ValueError("Argument", sub_arg, "not recognized.")
      elif opt == "-d":
        if len(opt) == 1:
          delim = opt
        elif len(opt) > 1:
          raise ValueError("Argument", sub_arg, "not recognized.")
      elif opt == "-s":
        arg = arg.split(",")
        for sub_arg in arg:
          try:
            index_substract.append(int(sub_arg))
          except:
            raise ValueError("Argument", sub_arg, "not recognized.")
      elif opt == "-a":
        arg = arg.split(",")
        for sub_arg in arg:
          try:
            index_add.append(int(sub_arg))
          except:
            raise ValueError("Argument", sub_arg, "not recognized.")
      elif opt == "-m":
        arg = arg.split(",")
        if len(arg) == 1:
          try:
            multiply = float(arg[0])
          except:
            raise ValueError("Argument", arg, "not recognized.")
        else:
          multiply = {}
          for sub_arg in arg[1:]:
            try:
              multiply[int(sub_arg)] = float(arg[0])
            except:
              raise ValueError("Argument", sub_arg, "not recognized.")
    
    if (index_data == "all" and (len(index_weights) > 0 
        or len(index_add) > 0 or len(index_substract) > 0)):
      raise ValueError("If using all columns for data,cannot specify "\
                         "weights, additions or substraction.")
    if len(index_data) != len(index_weights) and len(index_weights) > 0:
      raise ValueError("If specifying weights, must specify as many weights "\
                       "as data.")
      return
    for iw in index_weights:
      if iw in index_data:
        raise ValueError("Can not use column", iw,"for both data and weigths.")
    if len(index_weights) == 0:
      print(MyHist.skip, "Info in MyHist::load_file: Filling histogram using "\
                "column(s)", index_data)
    else:
      print(MyHist.skip, "Info in MyHist::load_file: Fillin histogram using"\
                  "  column(s)" , index_data, " for data and column(s)",
                 index_weights, "for weights.")
    if len(index_substract) > 0:
      print(MyHist.skip, "Info in MyHist::load_file: column(s)",
              index_substract, "will be substracted")
    if len(index_add) > 0:
      print(MyHist.skip, "Info in MyHist::load_file: column(s)", index_add,
                "will be added")
      
    data = []
    weights = []
    if not os.path.isfile(str("").join((path, filename))):
      raise OSError("File", str("").join((path, filename)), "not found.")
    with open(str("").join((path, filename)), "r") as ifile:
      lines = ifile.readlines()
      if ignore_first_line == True:
        lines = lines[1:]
      tb.clean_white_spaces(lines)
      for iline in range(len(lines)):
        line = lines[iline]
        line = line.split("\n")[0].split(delim)
        if ignore_first_column == True:
          line = line[1:]
        if index_data == "all":
          index_data = list(range(len(line)))
        if len(line) <= max(index_data + index_weights
                            + index_substract + index_add):
          warnings.warn(str(" ").join(("Line", str(iline), "(", str(line), 
                                       ") is ignored")), Warning)
          continue
        good_line = True
        tmp_data = []
        tmp_weights = []
        if multiply is not None:
          for ic in range(len(line)):
            try:
              line[ic] = float(line[ic])
            except:
              continue
            if isinstance(multiply, numbers.Number):
              line[ic] = line[ic]*multiply
            elif isinstance(multiply, dict):
              if ic in multiply:
                line[ic] = line[ic]*multiply[ic]
        for idw in range(len(index_data)):
          idata = index_data[idw]
          try:
            d = float(line[idata])
            if len(index_substract) > 0:
              for s in index_substract:
                d = d - float(line[s])
            if len(index_add) > 0:
              for a in index_substract:
                d = d + float(line[a])
            tmp_data.append(d)
            if len(index_weights) > 0:
              iw = index_weights[idw]
              w = float(line[iw])
              tmp_weights.append(w)
          except:
            good_line = False
            warnings.warn(str(" ").join(("Line", str(iline), "(", str(line), 
                                         ") is ignored")), Warning)
            continue
        if good_line:
          data += tmp_data
          if len(index_weights) > 0:
            weights += tmp_weights
    self.__data = data[:]
    if len(weights) > 0:
      self.__weights = weights[:]
    self.reinitialise()  # Force re-computation
    self._update()


# Overloads
  def __str__(self, all_bins = False):
    """ Will show statbox and fitboxes
    
    And bin content alonside with bin position if all_bins is true
    """
    
    if not self.is_up_to_date():
      self._update()
    to_return = self.__statbox.replace("\\\\", "\n")
    if len(self.__fitboxes) > 0:
      for f in self.__fitboxes:
        to_return = str("\n").join((to_return, f)).replace("\\\\", "\n")
    if all_bins:
      for c,m in zip(self.__bin_contents, self.__bin_middles):
        to_return = str("\n  ").join((
          to_return, str(" ").join((str(m), str(c)))
        ))
    return to_return

  def __getitem__(self, ibin):
    """get bin content based on bin number ibin."""
    
    return self.get_bin_content(ibin)

  def __eq__(self, other):
    """ Compares self with the instance of MyHist given as 'other'.

    False if other is None
    True if bin contents and bin edges are all equal
    and bin widths are equal and ranges are equal.
    """ 
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return False
    # Should be modified to tolerate floating point imprecision
    if self.__bin_width != other.get_bin_width():
      return False
    # Should be modified to tolerate floating point imprecision
    if self.__arange != other.get_range():
      return False
    if len(self.__bin_contents) != len(other.get_bin_contents()):
      return False
    if len(self.__bin_contents) == 0:
      return True
    for ic in range(len(self.__bin_contents)):
      c1 = self.__bin_contents[ic]
      e1 = self.__bin_edges[ic]
      c2 = other.get_bin_content(ic)
      e2 = other.get_bin_edge(ic)
      if c1 != c2 or e1 != e2:
        return False
    return True

  def __ne__(self,other):
    """ Compares self with the instance of MyHist given as 'other'.

    True if other is None
    False if bin contents are all equal
    """ 
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return True
    # Should be modified to tolerate floating point imprecision
    if self.__bin_width != other.get_bin_width():
      return True
    # Should be modified to tolerate floating point imprecision
    if self.__arange != other.get_range():
      return True
    if len(self.__bin_edges) != len(other.get_bin_edges()):
      return True
    for ic in range(len(self.__bin_contents)):
      c1 = self.__bin_contents[ic]
      e1 = self.__bin_edges[ic]
      c2 = other.get_bin_content(ic)
      e2 = other.get_bin_edge(ic)
      if c1 != c2 or e1 != e2:
        return True
    return False

  def __lt__(self,other):
    """ True if self has a smaller integral than other
    
    False if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return False
    return self.__integral < other.get_integral()

  def __le__(self,other):
    """ True if self has a smaller or equal integral than other
    
    False if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return False
    return self.__integral <= other.get_integral()

  def __gt__(self,other):
    """ True if self has a bigger integral than other 
    
    True if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return True
    return self.__integral > other.get_integral()

  def __ge__(self,other):
    """ True if self has a bigger or equal integral than other
    
    True if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return True
    return self.__integral >= other.get_integral()

  def __add__(h1, h2):
    """ Addition of histograms 'h1' and 'h2'. Must be MyHist instances.
    
    Will create a new histogram which will be the sum of the two input histos.
    The range will cover both input histos, the bin size will be the min of
    the bin sizes of the two input histos.  Does not use the data of the two
    input histos, but directly the bin contents.
    """
    
    if not h1.is_up_to_date():
      h1._update()
    if not h2.is_up_to_date():
      h2._update()
    newrange = [min(h1.get_range()[0], h2.get_range()[0]),
                 max(h1.get_range()[1], h2.get_range()[1])]
    bin_width = min(h1.get_bin_width(), h2.get_bin_width())
    no_number = h1.get_name_numbering() or h2.get_name_numbering()
    h = MyHist(name = str("").join((h1.get_name(), "+", h2.get_name())),
               binwidth=bin_width, no_name_numbering = no_number)
    for c, e in zip(h1.get_bin_contents(), h1.get_bin_edges()[:-1]):
      h.fill(e, c)
    for c, e in zip(h2.get_bin_contents(), h2.get_bin_edges()[:-1]):
      h.fill(e, c)
    h._update()
    return h

# Histogram computing and updating methods
  def reinitialise(self, total=True):
    """ Reinitialise (or create if histogram is new) some attributes.
    
    Used any time a creation method is used, or when set_data
    method is used, or just before computing the histogram.
    Can also just reinitialize the statistical attributes (mean, rms...)
    if argument 'total' is set to False (default True).
    """
    
    if total:
      self.__bin_width = -1
      self.__bin_edges = []
      self.__bin_middles = []
      
    self.__bin_contents = [0]*len(self.__bin_middles)  # do not use self_bins
                                                     # because can be 'auto'
    self.__underflow = 0
    self.__overflow = 0
    self.__mean = self.__rms = self.__integral = 0
    self.__count = 0
    self.__statbox = ""
    self.__fits = []
    self.__xyfits = []
    self.__fitted_params = []
    self.__fitboxes = []
    self.__chisquares = []
    self.__up_to_date = False

  def _prepare_computation(self):
    """ Determines if the histogram should be computed.
    
    See _compute_histo() for more details on the conditions.
    """
    
    # Histogram had already been computed.
    if len(self.__bin_edges) != 0:
      return False
    # Enough to use numpy.histogram.
    if len(self.__arange) != 0 and self.__bins != "auto":
      return True
    # Computes number of bins from range and specified bin width.
    if len(self.__arange) != 0 and self.__fixed_bin_width != -1:
      self._compute_number_of_bins()
      return True

    if len(self.__data) != 0:
      # If range was not specified, use min and max of data. Use a bit more
      # than the max to avoid overflow
      if len(self.__arange) == 0:
        self.__arange = [Decimal(min(self.__data)),
                        Decimal(max(self.__data)*1.0000001)]
      # Computes number of bins from range and specified bin width.
      if self.__fixed_bin_width != -1:
        self._compute_number_of_bins()
      return True
    
    return False

  def _compute_histo(self):
    """ Creates the numpy.histogram object.
    
    Called by the update method only if the histogram is new.  In other words,
    only of the attribute __bin_edges is an emtpy list.
    Will check that all attributes are compatible with one another and are ok.
    Computes the bin contents, edges and middles.
    Computes the number of overflow and underflow in the data set.
    If bins was set to auto, it will then be updated with the resulting number
    of bins.
    If range was set in decimal format, it will now be floating values.
    Data and weights are emptied upon computation.
    
    The histogram can be created completely empty.  In that case, nothing is
    to be done.  Empty means that there is no way to establish the range,
    number of bins and bin width of the x axis.
    So, the histogram is computed if:
      - The data set is not empty.  In that case:
        - If range was not given, it is assumed to be the min and max of the
          data set.
        - If bins was given as argument, it is used.
        - If binwidth was given as argument, computes bins from it and range.
      - Having an empty data set but specifying either bins and range or
        binwidth and range.
    """
    
    self.reinitialise(True)
    if not self._prepare_computation():
      return
    
    if len(self.__weights) != 0:
      if self.__bins == "auto":
        raise tb.ArgError("Cannot have auto bins and weighted bins at "\
                               "the same time.")
      if len(self.__data) != len(self.__weights):
        raise tb.ArgError("__weights and __data must have same size.")
        
    # Check if range makes sense.
    if self.__arange[0] > self.__arange[1]:
      tmp = self.__arange[0]
      self.__arange[0] = self.__arange[1]
      self.__arange[1] = tmp
    # Convert decimal range to float.
    tmp_range = []
    if len(self.__arange) > 0:
      tmp_range = [float(self.__arange[0]), float(self.__arange[1])]
    # Compute histo even if data is empty (bin edges will be up to date).
    if len(self.__weights) == 0:
      self.__weights = None
    self.__bin_contents, self.__bin_edges = np.histogram(
                                       self.__data,
                                       bins = self.__bins,
                                       range = tmp_range,
                                       density = self.__density,
                                       weights = self.__weights)
    self.__bin_contents = list(self.__bin_contents)
    self.__bin_edges = list(self.__bin_edges)
    self.__bin_width = self.__bin_edges[1] - self.__bin_edges[0]
    if self.__bins == "auto":
      self.__bins = len(self.__bin_contents)
    for d in self.__data:
      if d < self.__bin_edges[0]:
        self.__underflow += 1
      if d >= self.__bin_edges[-1]:
        self.__overflow += 1
    # Compute center positions of bins.and empty data and weights.
    self.__bin_middles = [ (self.__bin_edges[i-1]+self.__bin_edges[i]) / 2
                         for i in range(1, len(self.__bin_edges)) ]
    self.__data = []
    self.__weights = []

  def _update(self):
    """ Update the statistical attributes of the object.
    
    Execute _compute_histo() if histogram is empty. Computing histogram will
    reinitialise all statistical and histogram attributes.
    If not empty, will only update the statistical attributes.
    
    Runs at object creation and before any kind of display, get or fit.
    User should not have to use it as it is supposed to run when needed.
    """
    
    if len(self.__bin_edges) == 0:
      self._compute_histo()
    self.__up_to_date = True
    if len(self.__bin_edges) == 0:
      return
    
    # Computes the statistical attributes.
    self.__count = 0
    self.__mean = 0.
    sqmean = 0.
    self.__rms = 0.
    self.__integral = 0
    for content, middle in zip(self.__bin_contents, self.__bin_middles):
      if not content == content:  # Empty bins can be np.nan
        continue
      self.__mean = self.__mean + content*middle
      sqmean = sqmean + content*middle**2
      self.__count = self.__count + content  # Can be float.
      self.__integral += content*self.__bin_width
    if self.__count > 0:
      self.__mean = self.__mean/self.__count
      sqmean = sqmean/self.__count
      try:
        self.__rms = math.sqrt(sqmean - self.__mean**2)
      except ValueError:
        self.__rms = 0.
    self._form_stat_box()

  def _compute_number_of_bins(self):
    """ Compute number of bins based on range and fixed bin size.
    
    Used by _prepare_computation().
    """
    
    if self.__arange[0] > self.__arange[1]:
      self.__arange = [self.__arange[1], self.__arange[0]]
    self.__bins = round(
                      (self.__arange[1]-self.__arange[0])
                      / Decimal(self.__fixed_bin_width)
                     )
    if self.__bins <= 0:
      raise ValueError("Number of bins must be positive.")
    self.__up_to_date = False


# Set and add methods.
  def set_name(self, name):
    """ Sets the name of the instance to the given string. """
    
    MyHist.names.remove(self.__name)
    self.__name = name
    if not self.__no_name_numbering:
      i = 2
      tmp_name = self.__name
      while tmp_name in MyHist.names:
        tmp_name = self.__name
        tmp_name += "-"+str(i)
        i += 1
      self.__name = tmp_name
    MyHist.names.append(self.__name)

  def set_data(self, data, reinitialise = True):
    """ Set the data attribute and recomputes the histogram if asked.
    
    Arguments are:
      data -- the list to set __data to
      reinitialise -- If True, will reinitialise the histogram. (Default True)
    """
    
    self.__data = data[:]
    if reinitialise:
      self.reinitialise(True)  # Will force a new computation.
      self._update()
    else:
      self._up_to_date = False

  def set_number_of_bins(self, nbins):
    """ Set the 'bins' attribute to the integer or string given as argument.
    
    Only makes sense to do that is bin contents and edges have not been
    computed yet.  Will implement a rebin method for that purpose.
    Use 'auto' to reset the value.
    """
    
    if len(self.__bin_edges) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if self.__fixed_bin_width != -1 and nbins != "auto":
      raise tb.ArgError("can not set __bins if __fixed_bin_width was "\
                             "specified.")
    self.__bins = int(nbins)
    self.__up_to_date = False

  def set_range(self, arange):
    """ Set the 'arange' attribute.
    
    Argument must be a list of min and max range.  Can be string, if so will
    convert it to decimal for better precision.
    Only makes sense to do that is bin contents and edges have not been
    computed yet.  Will implement a specific method for that purpose.
    Use [] to reset the value.
    """
    
    if len(self.__bin_edges) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if len(arange) != 2 and len(arange) != 0:
      raise ValueError("Wrong format. arange must be list of two elements or"\
                       " an empty list.")
    if ((not type(arange[0]) is Decimal and
         not type(arange[0]) is int and
         not type(arange[0]) is str)):
      warnings.warn("You should use Decimal instead of a plain float for "\
                    "arange to avoid floating point imprecisions.", Warning)
    arange[0] = Decimal(arange[0])
    arange[1] = Decimal(arange[1])
    self.__arange = [min(arange), max(arange)]
    self.__up_to_date = False

  def set_weights(self, weights, reinitialise=True):
    """ Set the weights attribute and recomputes the histogram if asked.
    
    Arguments are:
      weights -- list of weights to set
      reinitialise -- If True, will reinitialise the histogram. (Default True)
    """
    
    self.__weights = weights
    if reinitialise:
      self.reinitialise(True)
      self._update()
    else:
      self.__up_to_date = False

  def set_bin_width(self, binwidth):
    """ Set the 'fixed_bins_size' attribute to the given argument.
    
    Can be string, if so will convert it to decimal for better precision.
    Only makes sense to do that if bin contents and edges have not been
    computed yet.  Will implement a rebin method for that purpose.
    Use -1 to reset the value.
    """
    
    if len(self.__bin_edges) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if self.__bins != "auto" and binwidth != -1:
      raise tb.ArgError("Can not set binwidth if __bins "\
                             "was specified.")
    if isinstance(binwidth, str):
      try:
        binwidth = Decimal(binwidth)
      except:
        raise ValueError("Can not convert", binwidth, "to decimals.")
    if not isinstance(binwidth, Decimal) and not isinstance(binwidth, int):
      warnings.warn("You should use Decimal instead of a plain float for "\
                    "binwidth to avoid floating point imprecisions.", Warning)
    self.__fixed_bin_width = binwidth
    self.__up_to_date = False

  def set_up_to_date(self):
    """ Tells the object it is not up to date.
    
    Not sure when this should be used. Probably never.
    """
    self.__up_to_date = True

  def fill(self, values, weights=None):
    """Add xs values with weights to the appropriate bin.
    
    Arguments are:
      values -- x-positions to Fill the histogram at. Can be number-like or
        iterable-like.
      weights -- weights to add on each x. Can be number-like or
        iterable-like.
    """
    
    if not isinstance(values, collections.Iterable):
      if isinstance(values, numbers.Number):
        values = [values]
      else:
        return  # Does not raise an error to allow for possible np.nan values
                # to be passed to Fill in a loop
    if len(values) == 0:
      return
    
    update_weights = True
    if not isinstance(weights, collections.Iterable):
      if weights is None:
        update_weights = False
        weights = [1]*len(values)
      elif isinstance(weights, numbers.Number):
        weights = [weights]
      else:
        return  # Does not raise an error to allow for possible np.nan values
                # to be passed to Fill in a loop
    if len(values) != len(weights):
      raise tb.ArgError("values and weights must have the same length.")
    # If histogram is new. In that case, append values to data and weights
    # to weights.
    if len(self.__bin_edges) == 0:
      self.__data += list(values)
      if update_weights:
        self.__weights += list(weights)
      self.__up_to_date = False
      return

    for v, w in zip(values, weights):
      ibin = self.get_bin(v)
      if ibin == "underflow":
        self.__underflow += w
      elif ibin == "overflow":
        self.__overflow += w
      else:
        if len(self.__fits) != 0:
          self.__fits = []# If bin content is modified, the current fits 
                          # are meaningless and are suppressed.
        self.add_bin_content(ibin, w)  # Will set __up_to_date to False

  def add_bin_content(self, ibin, weight=1):
    """Add a weight (can be float) to bin ibin.
    
    Histogram needs to have been computed first.
    """
    
    if len(self.__bin_edges) == 0 and len(self.__data) != 0:
      self._update()
    elif len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    if not self.__bin_contents[ibin] == self.__bin_contents[ibin]:  # is nan
      self.set_bin_content(ibin,weight)
    else:
      self.__bin_contents[ibin] += weight
    self.__up_to_date = False

  def set_bin_content(self, ibin, value):
    """Set a value to bin ibin.
    
    Histogram needs to have been computed first.
    """
    
    if len(self.__bin_edges) == 0 and len(self.__data) != 0:
      self._update()
    elif len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    self.__bin_contents[ibin] = value
    self.__up_to_date = False

  def set_normalise(self, density = True):
    """ Set the 'density' attribute to the given argument (Default True).
    
    Only makes sense to do that is histogram has not been computed yet.
    """
    
    if len(self.__bin_edges) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    self.__density = bool(density)
    self.__up_to_date = False

  def set_decimals(self, value):
    """ Set the number of decimals to be used to display histogram stats.
    
    Will update statbox and fitbox.
    """
    
    self.__ndecimals = int(value)
    if self.__statbox != "":
      self._form_stat_box()
    if len(self.__fitboxes) != 0:
      self._form_fit_box()


# get methods
  def get_id(self):
    return self.__id

  def get_name(self):
    return self.__name

  def get_name_numbering(self):
    return self.__no_name_numbering

  def get_data(self):
    return self.__data[:]

  def get_number_of_bins(self):
    if not self.__up_to_date:
      self._update()
    return self.__bins

  def get_range(self):
    if not self.__up_to_date:
      self._update()
    return self.__arange[:]

  def get_drange(self):
    if not self.__up_to_date:
      self._update()
    return abs(self.__arange[1]-self.__arange[0])

  def is_normalised(self):
    return self.__density

  def get_weights(self):
    return self.__weights[:]

  def get_ndecimals(self):
    return self.__ndecimals

  def get_fixed_bin_width(self):
    return self.__fixed_bin_width

  def is_up_to_date(self):
    return self.__up_to_date

  def get_bin_width(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_width

  def get_bin_edges(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_edges[:]

  def get_bin_middles(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_middles[:]

  def get_bin_contents(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_contents[:]
    
  def get_underflow(self):
    if not self.is_up_to_date():
      self._update()
    return self.__underflow
    
  def get_overflow(self):
    if not self.is_up_to_date():
      self._update()
    return self.__overflow

  def get_mean(self):
    if not self.is_up_to_date():
      self._update()
    return self.__mean
    
  def get_std_dev(self):
    if not self.is_up_to_date():
      self._update()
    return self.__rms
    
  def get_integral(self):
    if not self.is_up_to_date():
      self._update()
    return self.__integral
    
  def get_entries(self):
    if not self.is_up_to_date():
      self._update()
    return self.__count

  def get_statbox(self):
    if not self.is_up_to_date():
      self._update()
    return self.__statbox[:]

  def get_fits(self):
    return self.__fits[:]

  def get_xyfits(self):
    return self.__xyfits[:]

  def get_fitted_params(self):
    return self.__fitted_params[:]

  def get_fitboxes(self):
    return self.__fitboxes[:]

  def get_chisquares(self):
    return self.__chisquares[:]

  def get_bin(self, value):
    """ Returns the bin number of the given value.
    
    Histogram needs to have been computed first, but does not need to be
    up to date since the content of the bin if not used.
    """
    
    if (len(self.__bin_edges) == 0 and len(self.__data) != 0):
      self._update()
    elif len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    if value >= self.__bin_edges[-1]:
      return "overflow"
    if value < self.__bin_edges[0]:
      return "underflow"
    return int((Decimal(value)-self.__arange[0])/Decimal(self.__bin_width))

  def get_bin_content(self, ibin):
    """ Returns the content of bin ibin.
    
    Returns None if ibin is not in range of bin_contents.
    """
    
    if not self.is_up_to_date():
      self._update()
    if ibin < 0 or ibin > len(self.__bin_middles):
      return None
    return self.__bin_contents[ibin]

  def get_bin_edge(self, ibin):
    """ Returns the left edge of bin ibin.
    
    Returns None if ibin is not in range of bin_contents.
    """
    
    if not self.is_up_to_date():
      self._update()
    if ibin < 0 or ibin > len(self.__bin_middles):
      return None
    return self.__bin_edges[ibin]

  def get_bin_middle(self, ibin):
    """ Returns the middle of bin ibin.
    
    Returns None if ibin is not in range of bin_contents.
    """
    
    if not self.is_up_to_date():
      self._update()
    if ibin < 0 or ibin > len(self.__bin_middles):
      return None
    return self.__bin_middles[ibin]

  def get_x_content(self, x):
    """ Returns the content of the bin corresponding to the given x.
    
    Returns None if x is not in range of histo
    """
    
    if not self.is_up_to_date():
      self._update()
    ibin = self.get_bin(x)
    if ibin < 0 or ibin > len(self.__bin_middles):
      return None
    return self.__bin_contents[ibin]

  def get_maximum(self):
    """ Returns the maximum count.
    
    Histogram needs to have been computed first and up to date.
    """
    
    if not self.is_up_to_date():
      self._update()
    elif len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    return max(self.__bin_contents)

  def get_position_of_maximum(self):
    """ Returns the position of the maximum count.
    
    Histogram needs to have been computed first and up to date.
    """
    
    if not self.is_up_to_date():
      self._update()
    elif len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    return self.__bin_middles[self.__bin_contents.index(self.get_maximum())]

# Fitting methods 
  def fit(self, method, params, arange=[], par_range=[]):
    """ Fits the histogram with the specified method.
    
    Uses scipy.optimise.leastsq method.  Will fill fits, xyfits, fitboxes,
    chisquare and fitted_params attribute.
    
    Arguments are:
      method -- any string readable by eval.
        Example: 'p[0]*math.exp(-p[1]*x)'.  The p[i] correspond to the
        parameters that will be fitted.  Do not forget to specify the math
        module!
        Does not suport passing full methods yet, but I intend to make it so.
      params -- the starting values of the parameters of the fit.
      arange -- the range on which to fit.  If not specified, the full range
        of the histo will be used.  Range can not be greater than histo range.
      par_range -- specifies the limits of the fit parameters (default [])
     """
     
    if not self.is_up_to_date():
      self._update()
    if len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    if len(arange) != 2:
      arange = self.__arange
    br = [self.get_bin(arange[0]), self.get_bin(arange[1])]
    if isinstance(br[0], str):
      br[0] = 0
    if isinstance(br[1], str):
      br[1] = self.__bins-1
    if len(params) == 0:
      raise tb.ArgError("Must specify starting values for parameters.")
    if isinstance(method, str):
      method2 = method[:]
      iclim = 0
      # Finds the positions of the parameters in the string.
      indicesopen = [i for i, a in enumerate(method2[:-1])
                       if a+method[i+1] == "p["]
      indicesclose = [i for i, a in enumerate(method2)
                        if a == "]"]
      if len(indicesopen) != len(indicesclose):
        raise ValueError("Can not convert '", method, "'to mathematical "\
                         "expression. Mismatch of p[ and ].")
      # Check if number of parameters matches the number of initial values.
      method_params = [method[i+2:j] for i, j 
                        in zip(indicesopen, indicesclose)]
      number_of_params = len(list(set(method_params)))
      if number_of_params != len(params):
        raise tb.ArgError("Numbers of parameters provided does not "\
                               "match number of parameters required.")
      for io, ic in zip(indicesopen, indicesclose):
        if io > ic:
          raise ValueError("Can not convert '", method, "' to mathematical "\
                           "expression. Mismatch of p[ and ].")
      # Checks if the string can be read by eval with dummy x=1 value.
      try:
        tb.function_from_string(params, 1, method2)
      except (ZeroDivisionError, ValueError):
        pass
      except:
        raise ValueError("Can not convert '", method, "' to mathematical "\
                         "expression. Tried with '", method2,"'. Make sure "\
                         "you used 'x' for the variable and that "\
                         "you put 'math.' before any function uses by math.")
      print(MyHist.skip, "Info in MyHist::fit: fitting histogram with function",
            method, ".")
    else:  # Tried to pass method isntead of string as argument.
      print("    Not implemented yet.")
      return
    # Does the fit and fill attributes
    fitted_params, chisquare, ffit = tb.general_fit(method,
                            self.__bin_middles[br[0]:br[1]],
                            self.__bin_contents[br[0]:br[1]],
                            params, par_range)
    self.__fits.append(method)
    self.__fitted_params.append(fitted_params)
    self.__chisquares.append(chisquare)
    self.__xyfits.append(ffit)
    self._form_fit_box()

  def gauss_fit(self, method="", arange=[], params=[]):
    """ Fitting method using gaussian
    
    Different from generic fit method to allow for not specifying
    the params, making it easier to fit the histo in a script.
    Uses scipy.optimise.leastsq.
    
    Arguments qre:
      method -- can be either 'single' or 'double'.  Double means
        the fit expects to find two distinct maxima and will fit two
        independant gaussians around them.  Could technically be done with N 
        maxima, but it is safer to use the generic fit method in that case.
    arange -- the fit range.  If not specified, will cover the full
      histogram range.  Can not be greater than the histogram range.
    params --the initial values of the fit parameters (normalisation,
      mean and rms, times 2 if method is double).
    
    If the params were not specified, Will use scipy.signal.find_peaks_cwt to
    locate the maxima:
    -If one max is found, will fit one gaussian, even is method is 'double'.
    -If two or more are found, will fit 2 gaussians, unless method is
    'single', then it will fit one gaussian on the greatest maximum.
    """
    
    if not self.is_up_to_date():
      self._update()
    if len(self.__bin_edges) == 0:
      raise ValueError("Histogram is empty.")
    if len(arange) != 2:
      arange = self.__arange
    else:
      arange = [Decimal(a) for a in arange]
    br = [self.get_bin(arange[0]), self.get_bin(arange[1])]
    if isinstance(br[0], str):
      br[0] = 0
    if isinstance(br[1], str):
      br[1] = self.__bins-1
    # Will try to find the maxima of the histogram if no parameters were
    # specified.
    if len(params) == 0:
      n_maxs = -1
      if method == "single":
        n_maxs = 1
      elif method == "double":
        n_maxs = 2
      elif method != "":
        raise ValueError("Method not recognized.")
      xmaxs, ymaxs = tb.find_local_max(self.__bin_middles[br[0]:br[1]],
                                       self.__bin_contents[br[0]:br[1]],
                                       arange, self.__bin_width, n_maxs)
      if len(xmaxs) == 0:
        warnings.warn("No maximum found. Fit failed", Warning)
        return
      elif len(xmaxs) == 1:
        # If method not secified, set method to single when we have one max.
        if method == "":
          method = "single"
        # If one max was found but we were trying to do double fit, retry to
        # find the maxima without specifying the number of expected max.
        elif method == "double":
          xmaxs2, ymaxs2 = tb.find_local_max(self.__bin_middles[br[0]:br[1]],
                                           self.__bin_contents[br[0]:br[1]],
                                           arange, self.__bin_width)
          # If we still have only one max, set the fitting method to single.
          if len(xmaxs2) == 1:
            warnings.warn("Found only one maximum. Will use single gauss fit"\
                          " instead of double.", Warning)
            method = "single"
          elif len(xmaxs2) == 2:
            xmaxs = xmaxs2
            ymaxs = ymaxs2
      # If method was not secified, set method to double when we have 2 max.
      elif len(xmaxs) == 2 and method == "":
        method = "double"
      else:
        warnings.warn("More than 2 maxima were found.", Warning)
        # If method was not specified, will try double gauss fit.
        if method == "":
          warnings.warn("Trying double gauss fit.", Warning)
          method = "double"
      # Compute the initial fit parameters.
      if method == "single":
        ymax = max(ymaxs)
        xmax = xmaxs[ymaxs.index(ymax)]
        params = [self.__integral,
                  xmax,
                  self.__rms]
      elif method == "double":
        ymax0 = max(ymaxs)
        i0 = ymaxs.index(ymax0)
        xmax0 = xmaxs[i0]
        ymaxs_copy = ymaxs[:]
        ymaxs_copy[i0] = min(ymaxs)
        ymax1 = max(ymaxs_copy)
        i1 = ymaxs.index(ymax1)
        xmax1 = xmaxs[i1]
        ratio = ymax0/ymax1
        int0 = ratio*self.__integral/(ratio+1)
        sigma0 = ratio*self.__rms/(ratio+1)
        int1 = self.__integral/(ratio+1)
        sigma1 = self.__rms/(ratio+1)
        params = [int0, xmax0, sigma0,
                  int1, xmax1, sigma1]
    # Do the fit based on specified method.
    if len(params) == 3:
      fitted_params, chisquare, ffit = tb.general_fit(
                                   "gauss",
                                   self.__bin_middles[br[0]:br[1]],
                                   self.__bin_contents[br[0]:br[1]],
                                   params)
      self.__fits.append("gauss")
      self.__fitted_params.append(fitted_params)
      self.__chisquares.append(chisquare)
      self.__xyfits.append(ffit)
      self._form_fit_box()
    elif len(params) == 6:
      fitted_params, chisquare, ffit = tb.general_fit(
                                   "doublegauss",
                                   self.__bin_middles[br[0]:br[1]],
                                   self.__bin_contents[br[0]:br[1]],
                                   params)
      self.__fits.append("doublegauss")
      self.__fitted_params.append(fitted_params)
      self.__chisquares.append(chisquare)
      self.__xyfits.append(ffit)
      self._form_fit_box()
    else:
      raise tb.ArgError("wrong number of parameters (need 3 for single "\
                             "gauss fit, 6 for double).")


# Display methods.
  def _form_stat_box(self):
    if not self.is_up_to_date():
      self._update()
    if self.__count is None:
      return None
    Entries = self.__count
    if Entries != int(Entries):
      Entries = tb.stround(Entries, self.__ndecimals)
    else:
      Entries = str(int(Entries))
    to_return = str("").join(("\\flushleft -------------- \\\\ \\textbf{", self.__name , ":"))
    to_return = str("\\\\ ").join((
                   to_return,
                   str(" ").join(("count:", Entries)),
                   str(" ").join(("mean:", tb.stround(self.__mean,
                                                      self.__ndecimals))),
                   str(" ").join(("$\\sigma$:", tb.stround(
                                                      self.__rms,
                                                      self.__ndecimals))),
                   str(" ").join(("$\\int$:", tb.stround(
                                                      self.__integral,
                                                      self.__ndecimals)))
                 ))
    if self.__underflow > 0:
      to_return = str("\\\\ ").join((
                     to_return,
                     str(" ").join(("Underflow:", str(self.__underflow)))
                   ))
    if self.__overflow > 0:
      to_return = str("\\\\ ").join((
                     to_return,
                     str(" ").join(("Overflow:", str(self.__overflow)))
                   ))
    self.__statbox = str("").join((to_return,"}", "\\\\ --------------"))

  def _form_fit_box(self):
    if len(self.__fits) == 0:
      return
    if len(self.__fitboxes) == 0:
      self.__fitboxes.append(str("").join(("\\flushleft -------------- \\\\ \\textbf{", self.__name, ":")))
    if self.__fits[-1] == "gauss":
      fitstring = " Fitted By:\\\\gauss"
      fitstring = str("\\\\  ").join((
                 fitstring,
                 str("=").join(("Int",
                   tb.stround(self.__fitted_params[-1][0],
                              self.__ndecimals)
                 )),
                 str("=").join(("mean",
                   tb.stround(self.__fitted_params[-1][1],
                              self.__ndecimals)
                 )),
                 str("=").join(("sigma",
                   tb.stround(self.__fitted_params[-1][2],
                              self.__ndecimals)
                 )),
                 str("=").join(("$\chi^2$",
                   tb.stround(self.__chisquares[-1],
                              self.__ndecimals)
                 ))
               ))
      fitstring = str("").join((fitstring, "\\\\ --------------"))
      
      self.__fitboxes[-1] = str("").join((
        self.__fitboxes[-1], fitstring, "}"
      ))
    elif self.__fits[-1] == "doublegauss":
      fitstring = " Fitted By:\\\\doublegauss"
      fitstring = str("\\\\  ").join((
                                          fitstring,
                                          "first gaussian:"
                                          ))
      fitstring = str("\\\\    ").join((
                 fitstring,
                 str("=").join(("Int",
                   tb.stround(self.__fitted_params[-1][0],
                              self.__ndecimals)
                 )),
                 str("=").join(("mean",
                   tb.stround(self.__fitted_params[-1][1],
                              self.__ndecimals)
                 )),
                 str("=").join(("sigma",
                   tb.stround(self.__fitted_params[-1][2],
                              self.__ndecimals)
                 ))
               ))
      fitstring = str("\\\\  ").join((
                                          fitstring,
                                          "second gaussian:"
                                          ))
      fitstring = str("\\\\    ").join((
                 fitstring,
                 str("=").join(("Int",
                   tb.stround(self.__fitted_params[-1][3],
                              self.__ndecimals)
                 )),
                 str("=").join(("mean",
                   tb.stround(self.__fitted_params[-1][4],
                              self.__ndecimals)
                 )),
                 str("=").join(("sigma",
                   tb.stround(self.__fitted_params[-1][5],
                              self.__ndecimals)
                 ))
               ))
      fitstring = str("\\\\  ").join((
                       fitstring,
                       str("=").join(("$\chi^2$",
                       tb.stround(self.__chisquares[-1],
                                  self.__ndecimals)))
                     ))
      fitstring = str("").join((fitstring, "\\\\ --------------"))
      
      self.__fitboxes[-1] = str("").join((
        self.__fitboxes[-1], fitstring, "}"
      ))
    else:
      fitstring = str("\\\\").join((" Fitted By:",self.__fits[-1]))
      for ipar in range(len(self.__fitted_params[-1])):
        fitstring = str("\\\\  ").join((
              fitstring,
              str("").join(("p[",str(ipar),"]: ",
                tb.stround(self.__fitted_params[-1][ipar], 
                self.__ndecimals)
              ))
            ))
      fitstring = str("\\\\  ").join((
              fitstring,
              str("=").join(("$\chi^2$",
                tb.stround(self.__chisquares[-1],
                self.__ndecimals)
              ))
            ))
      fitstring = str("").join((fitstring, "\\\\ --------------"))
      
      self.__fitboxes[-1] = str("").join((
        self.__fitboxes[-1], fitstring, "}"
      ))

  def save(self, filename="", path = ""):
    """ Save the histogram to a file, readable with read method """
    
    if path != "":
      if not os.path.isdir(path):
        raise OSError("Directory",path,"not found.")
    if not self.is_up_to_date():
      self._update()
    if filename == "":
      filename = str("").join(("MyHist_", self.__name, ".hdf5"))
    fullname = str("").join((path, filename))
    if "/tmp" not in fullname:
      print(MyHist.skip, "Info in MyHist::save: Saving histogram in file",
            fullname)
    with h5py.File(fullname,"w") as ofile:
      ofile.create_dataset("name", data=[np.string_(self.__name)])
      ofile.create_dataset("id_ndecimals",
                            data=[self.__id, self.__ndecimals], dtype='i8')
      if isinstance(self.__bins, str):
        ofile.create_dataset("bins_str", data=[np.string_(self.__bins)])
      else:
        ofile.create_dataset("bins_int", data=[self.__bins], dtype='i8')
      ofile.create_dataset("bin_width", data=[self.__bin_width], dtype='f8')
      if isinstance(self.__fixed_bin_width, Decimal):
        ofile.create_dataset("fixed_bin_width_str",
                              data=[np.string_(self.__fixed_bin_width)])
      else:
        ofile.create_dataset("fixed_bin_width_f",
                             data=[self.__fixed_bin_width], dtype='f8')
      ofile.create_dataset("statistics", data=[self.__mean,
                                              self.__rms,
                                              self.__integral,
                                              self.__underflow,
                                              self.__overflow,
                                              self.__count], dtype='f8')
      ofile.create_dataset("statbox",
                           data=[np.string_(self.__statbox)])
      ofile.create_dataset("bools", data=[self.__no_name_numbering,
                                          self.__up_to_date,
                                          self.__density], dtype='?')
      ofile.create_dataset("data", data=self.__data[:], dtype='f8')
      if isinstance(self.__arange[0], Decimal):
        ofile.create_dataset("arange_str", data=[np.string_(a)
                                             for a in self.__arange])
      else:
        ofile.create_dataset("arange_float",
             data=self.__arange[:], dtype='f8')
      ofile.create_dataset("weights", data=self.__weights[:], dtype='f8')
      ofile.create_dataset("bin_edges", data=self.__bin_edges[:], dtype='f8')
      ofile.create_dataset("bin_middles",
           data=self.__bin_middles[:], dtype='f8')
      ofile.create_dataset("bin_contents",
           data=self.__bin_contents[:], dtype='f8')
      ofile.create_dataset("fits", data=[np.string_(f) for f in self.__fits])
      ofile.create_dataset("xyfits", data=self.__xyfits[:], dtype='f8')
      ofile.create_dataset("fitted_params",
           data=self.__fitted_params[:], dtype='f8')
      ofile.create_dataset("fitboxes", data=[np.string_(f)
                                             for f in self.__fitboxes])
      ofile.create_dataset("chisquares",
           data=self.__chisquares[:], dtype='f8')

  def plot(self, **kwargs):
    """ Method to plot and/or save the histogram as a bar chart.
    
    All arguments are optional.
    Arguments are:
      (pltpdf)pdf -- will save the plot as a new page in the given pdf.
        Can not be specified along with the ax argument. (default None)
      (plt.subplot)ax -- will plot the histogram on the given subplot.
        The limits of the given subplot will be updated to fit the range
        of the new histogram.
        If not specified, a new subplot will be created.
        If pdf was not specified either, the subplot (new or not) is returned
        by the method.
      (strings)xlabel, ylabel and title -- self explanatory.
      (string)option -- if contains 'save', will save the plot in a pdf file
        named 'MyHist_self.__name.pdf' (independently from the pdf argument).
        If it contains 'show', will show the plot on screen.  Both can be used
        at the same time, either by writing 'save show', 'show save', or
        without spaces. (default '')
      (string)fitcolor -- color to use to plot any fitted function. If not
        specified, will be same color as the bar chart if fitted.
      -(list)x/yticks -- list of ticks to use on x/y axis. (default None)
      -(string)x/yticksformat: used to specify format of x/y tick labels.
        (default None)
      -(float)x/yticksfontsize -- self explanatory
      -(bool)force_xlim -- will force the range of x axis to match the range
        of the histogram, +/- 1%
      -(bool) adapt_x -- if ax is specified and the current limits
        do not allow to display the entire histo, will change them
        so they do (default False).
      -(bool) hide_statbox/fitbox -- self explanatory
      -(string) label -- label to give to the image, to be used later in legend.
        (default is statbox, fitbox or None)
    
    If no pdf or ax is specified, will return a new plt.subplot ax containing
    the plot of the histogram, with all its fitted functions.
    """
    
    expected = ["pdf", "ax", "c", "ec", "lw", "xlabel", "ylabel", 
                "xlabelfontsize", "ylabelfontsize",
                "title", "titlefontsize", "option", "fitcolor",
                "xticks", "yticks", "xticksformat", "yticksformat",
                "xticksspacing", "yticksspacing",
                "xticksfontsize", "yticksfontsize",
                "force_xlim", "adapt_xlim", "xlim", "ylim",
                "legendfontsize", "hide_statbox", "hide_fitbox", "path", "label"]
    tb.check_kwargs(expected, kwargs)
    
    if not self.__up_to_date:
      self._update()
    pdf = kwargs.get("pdf", None)
    ax = kwargs.get("ax", None)
    c = kwargs.get("c", None)
    ec = kwargs.get("ec", "grey")
    lw = kwargs.get("lw", None)
    xlabel = kwargs.get("xlabel", "\\textbf{x}")
    ylabel = kwargs.get("ylabel", str("").join(("\\textbf{Count/",
                                  tb.stround(self.__bin_width,
                                  self.__ndecimals), "}")))
    xlabelfontsize = kwargs.get("xlabelfontsize", 20)
    ylabelfontsize = kwargs.get("ylabelfontsize", 20)
    title = kwargs.get("title", None)
    titlefontsize = kwargs.get("titlefontsize", 20)
    option = kwargs.get("option", "")
    fitcolor = kwargs.get("fitcolor", None)
    xticks = kwargs.get("xticks", None)
    yticks = kwargs.get("yticks", None)
    xticksformat = kwargs.get("xticksformat", None)
    yticksformat = kwargs.get("yticksformat", None)
    xticksspacing = kwargs.get("xticksspacing", None)
    yticksspacing = kwargs.get("yticksspacing", None)
    xticksfontsize = kwargs.get("xticksfontsize", 15)
    yticksfontsize = kwargs.get("yticksfontsize", 15)
    force_xlim = kwargs.get("force_xlim", False)
    adapt_xlim = kwargs.get("adapt_xlim", False)
    xlim = kwargs.get("xlim", [])
    ylim = kwargs.get("ylim", [])
    legendfontsize = kwargs.get("legendfontsize", 15)
    hide_statbox = kwargs.get("hide_statbox", False)
    hide_fitbox = kwargs.get("hide_fitbox", False)
    path = kwargs.get("path", "")
    label = kwargs.get("label", None)
    if pdf is not None and ax is not None:
      raise tb.ArgError("Can not specify a pdf to save to and a subplot"\
                             " to update at the same time.")
    fig = None
    ax1 = ax
    if pdf is not None:
      plt.close("all")
      fig = plt.figure(0)
    if ax1 is None:
      if pdf is None:
        plt.close("all")
      ax1 = plt.subplot(111)
    else:
      # Get the current x limits if ax specified. If adapt_xlim is,
      # specified check if the current limit is too high or too low to display
      # the entire plot and change it if need be.
      xlimlow, xlimhigh = ax1.get_xlim()
      if Decimal(xlimlow) > self.__arange[0] and adapt_xlim:
        ax1.set_xlim(left = float(self.__arange[0]))
      if Decimal(xlimhigh) < self.__arange[1] and adapt_xlim:
        ax1.set_xlim(right = float(self.__arange[1]))
    
    
    # If xlim is specified, set it. Otherwise, if force_xlim is specified,
    # redefine the x limites to match the histo ranges.
    if len(xlim) == 2:
      ax1.set_xlim(left=xlim[0], right=xlim[1])
    elif force_xlim:
      ax1.set_xlim([0.99*float(self.__arange[0]),
                    1.01*float(self.__arange[1])])
    if len(self.__fits) == 0:
      if label is None:
        label = self.__statbox if not hide_statbox else None
      ax1.bar(self.__bin_edges[:-1],
              self.__bin_contents,
              self.__bin_width,
              color = c,
              edgecolor = ec,
              linewidth = lw,
              align="edge",
              label = label)
    else:
      myplot = ax1.bar(self.__bin_edges[:-1],
              self.__bin_contents,
              self.__bin_width,
              color = c,
              edgecolor = ec,
              linewidth = lw,
              align="edge",
              alpha = 0.75,
              label = label)
      for i in range(len(self.__fits)):
        tmp_fitcolor = fitcolor
        if fitcolor is None:
          tmp_fitcolor = myplot.patches[0].get_facecolor()
        ax1.plot(self.__xyfits[i][0],
                 self.__xyfits[i][1],
                 label = self.__fitboxes[i] if not hide_fitbox else None,
                 c = tmp_fitcolor)
    if ylabel is not None:
      ax1.set_ylabel(ylabel, fontsize=ylabelfontsize)
    if xlabel is not None:
      ax1.set_xlabel(xlabel, fontsize=xlabelfontsize)
    if title is not None:
      ax1.set_title(title, fontsize = titlefontsize)
    
    if xticksfontsize is not None:
      ax1.tick_params(axis='x', labelsize=xticksfontsize)
    if yticksfontsize is not None:
      ax1.tick_params(axis='y', labelsize=yticksfontsize)

    if xticks is not None:
      if xticks == 'bins':
        xticks = self.__bin_edges
      ax1.set_xticks(xticks)
    if yticks is not None:
      ax1.set_yticks(yticks)
    if xticksspacing is not None:
      ax1.xaxis.set_major_locator(ticker.MultipleLocator(xticksspacing))
    if yticksspacing is not None:
      ax1.yaxis.set_major_locator(ticker.MultipleLocator(yticksspacing))
    if xticksformat is not None:
      ax1.xaxis.set_major_formatter(FormatStrFormatter(xticksformat))
    if yticksformat is not None:
      ax1.yaxis.set_major_formatter(FormatStrFormatter(yticksformat))
      
    if ((len(self.__fitboxes) != 0 and not hide_fitbox)
        or (self.__statbox != "" and not hide_statbox)):
      leg = ax1.legend(bbox_to_anchor=(1.01, 0, 0, 1),
                       loc="upper left",
                       facecolor="wheat",
                       fontsize = legendfontsize,
                       shadow=True,
                       title = None)
       
      plt.gcf().canvas.draw()
      renderer = plt.gcf().canvas.renderer
      # Returns [[x0,y0],[x1,y1]]
      ax_xy0, ax_xy1 = ax1.get_window_extent(renderer).get_points()
      leg_xy0, leg_xy1 = leg._legend_box.get_window_extent(
                                      renderer).get_points()
      att = 0
      # Adjust the legend box to fit the subplot
      # If y0 of leg is smaller than that of ax, or y1 is higher than ax
      while (ax_xy0[1] > leg_xy0[1] or ax_xy1[1] < leg_xy1[1]) and att < 100:
        att += 1
        fontsize = leg._fontsize*0.85
        leg = ax1.legend(bbox_to_anchor=(1.01, 0, 0, 1),
                         loc="upper left",
                         facecolor="wheat",
                         shadow=True,
                         title = None,
                         fontsize = fontsize)
        ax_xy0, ax_xy1 = ax1.get_window_extent(renderer).get_points()
        leg_xy0, leg_xy1 = leg._legend_box.get_window_extent(
                                        renderer).get_points()
        legx1.append(leg_xy1[0])
        legy1.append(leg_xy1[1])
        axx1.append(ax_xy1[0])
        axy1.append(ax_xy1[1])
      if att == 100:
        warnings.warn("Legend might be too big", Warning)
    plt.gcf().tight_layout()
    
    if "save" in option and pdf is None:
      path = os.path.join(
          path,
          str("").join(("MyHist_", self.__name, ".pdf"))
        )
      plt.savefig(path, bbox_inches="tight")
    if "show" in option:
      plt.show()
    if pdf is not None:
      pdf.savefig(fig, bbox_inches="tight")
      plt.close("all")
    else:
      return ax1
