""" MyHist2D object

Created by Philippe COTTE, cottephi@gmail.com, https://gitlab.com/cottephi

Requieres toolbox module (part of the same gitlab project)

Uses numpy.histogram objects to compute bin contents and edges from a list of
data, with the possibility to specify weights for each entry. Weights will
be considered as seperate entries.
Has numerous methods allowing to update the bin contents later on, based on
the TH1 histogram of Root (CERN software).
Has a build-in plot function printing all relevent information of the histo
(mean, rms, overflow, underflow, counts, integral, fits if any...)
Has a built-in method to fit the histogram with one or two gaussians, or with
a user-specified method (uses eval to read a string.  Will be improved to
accept an full method as argument).
Can be saved to hdf5 file that can be re-read later to recreate the
histogram object in a different script.
Can read data from simple text file, supporting a lot of options to choose
which columns to read, add or substract.

  Class attributes:
    Created in all creation methods: 
      __id -- integer unique to each instance of the class.
      __name -- name of the object.
      __data -- raw data used to compute the histogram. Must be an iterable of
        format [x,y] where x and y are two iterables of same size.
      __bins_x/y -- number of bins along x/y axis. Can be 'auto'.
      __arange_x/y  -- x/y-axis range.
      __density -- bool saying whether the histogram has been normalised.
      __weights -- weights to apply to the raw data.
      __ndecimals -- precisions of printed numerical attributes.
      __fixed_bin_width_x/y -- bin width of x/y axis specified by the user.
      __up_to_date -- says whether the statistical attritbutes are up to date.
      
    Created in reinitialise, read and copy methods:
      Histogram attributes:
        __bin_width_x/x -- actual bin width on x/y axis, computed from
          bin_edges. Can be slightly different from fixed_bin_width because of
          roundings.
        __bin_edges_x/y -- the x/y position of the left side of each bin, plus
          the x/y position of the right side of the last bin
          (length is len(bin_middles) + 1)).
        __bin_middles_x/y -- the x/y position of the middle of each bin.
        __bin_contents -- counts in each bin.
          Format is [[c,...,c], ... , [c,...,c]] where size of the outer list
          is the number of x bins and the sizes of the inner lists are all
          equal to the number of y bins.
        __underflow -- number of entries below range.
        __overflow -- number of entries above range.
      Statistical attributes:
        __mean_x/y -- mean of the projected distribution of the x/y axis.
        __rms_x/y -- standard deviation of the above-mentionned distribution.
        __integral -- sum of content*bin_width_x*bin_width_y.
        __count -- total number of entries (can be float due to weights).
        __statbox -- string used to display the statistical attributes.
      Fit attributes:
        __fits -- list of all fits that were done (strings containing
          either the name of the method used or the methmatical expression).
        __xyzfits -- list of list of pairs (x,y) of the fitted function.
        __fitted_params -- list of list of parameters for each fit.
        __fitboxes -- list of strings used to dispay the fit results in termes
          of paramters value.
        __chisquares -- list of chisquares of each fit.
      NOTE : fitting function is not implemented yet.
  
  Public Methods:
    All getters return a copy of the class attribute unless specified
    otherwise.
    Not written yet
"""

import warnings
import numpy as np
import os
import warnings
import math
import getopt
import collections
import numbers
import cycler
import h5py
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.backends.backend_pdf as pltpdf
import matplotlib.gridspec as gridspec
from matplotlib import ticker
from decimal import Decimal
from copy import copy
from copy import deepcopy

import toolbox as tb

class MyHist2D:
  # Color blind-friendly color cycle
  CB_color_cycle = cycler.cycler(color=["#377eb8", "#ff7f00", "#4daf4a",
                                        "#f781bf", "#a65628", "#984ea3",
                                        "#999999", "#e41a1c", "#dede00"])
  mpl.rcParams["axes.prop_cycle"] = CB_color_cycle

  # Count the number of class instances created and store their names
  names = []
  skip = ""

  @classmethod
  def get_instances(cls):
    return MyHist2D.names[:]

  @classmethod
  def clear(cls):
    MyHist2D.names = []

  @classmethod
  def set_skip(cls, skip):
    MyHist2D.skip = skip

  def __init__(self, **kwargs):
    """ Class creator.
    
    Keyword arguments:
    filename -- a string that is passed to the read() method to create the
      histogram from a hdf5 file containing a previously saved histogram. Only
      compatible with 'name'. (default '')
    name -- a string, name of this instance of the class.  Will be appended
      with _integer if the name was already taken by another instance of the
      class. (default is the current number of instances of the class)
    data -- list of data to use to plot the histogram. (default [[],[]])
    bins_x/y -- the number of bins to use on x/y. Not compatible with
      'binwidth'. (Defalt -1). No automatic binning.
      Mandatory if binwidth not specified.
    arange_x/y -- Range of the histogram's x/y axis.  It is a 2-elements list.
      Supports the use of string that will be converted to decimals,
      which is recommanded to avoid floating point imprecisions. Default will
      be the min and max of 'data[0]' (xs) and 'data[1]' (ys) at histogram
      computation.
    density -- boolean to tell if the histogram is to be normalised.
      (default False)
    weights -- will apply a weight to each entry in 'data'.  Must have the
      same length than 'data[0]' and 'data[1]'.  Can not be used along with
      automatic binning. (default [])
    ndecimals -- integer, determines the precision for stat and fit boxes.
      (default 2)
    binwidth_x/y -- will force the bin width on x/y axis. Not compatible with
    'bins'. (default -1). Mandatory if bins_x/y not specified.
    zero_to_nan -- will replace all zeros with nans.
    
    """
    
    expected = ["filename", "name", "data", "bins_x", "bins_y", "arange_x",
                "arange_y", "ndecimals", "binwidth_x", "binwidth_y",
                "weights", "density","zero_to_nan"]
    tb.check_kwargs(expected, kwargs)
    
    filename = kwargs.get("filename", "")
    self.__name = kwargs.get("name", "")
    i = 1
    tmp_name = self.__name
    while tmp_name in MyHist2D.names:
      tmp_name = self.__name
      tmp_name += "_"+str(i)
      i += 1
    self.__name = tmp_name
    
    if filename != "":
      self.read(filename)
      return
    
    self.__id = len(MyHist2D.names)
    if self.__name == "":
      self.__name = str(self.__id)
    self.__data = list(kwargs.get("data", [[],[]]))
    if self.__data == []:
      self.__data = [[],[]]
    if len(self.__data) != 2:
      raise ValueError("data format incorrect.")
    if len(self.__data[0]) != len(self.__data[1]):
      raise ValueError("data format incorrect.")
  
    self.__bins_x = kwargs.get("bins_x", -1)
    if not isinstance(self.__bins_x, int):
      raise ValueError("bins_x must be int.")
    
    self.__bins_y = kwargs.get("bins_y", -1)
    if not isinstance(self.__bins_y, int):
      raise ValueError("bins_y must be int.")
    
    self.__arange_x = kwargs.get("arange_x", [])
    if len(self.__arange_x) == 2:
      try:
        self.__arange_x[0] = Decimal(self.__arange_x[0])
        self.__arange_x[1] = Decimal(self.__arange_x[1])
      except:
        raise ValueError("Wrong arange_x format. Must be convertible to "\
                         "Decimal.")
    elif len(self.__arange_x) != 0:
      raise ValueError("Wrong arange_x format. Must be a list"\
                      " of two elements or an empty list.")
    
    self.__arange_y = kwargs.get("arange_y", [])
    if len(self.__arange_y) == 2:
      try:
        self.__arange_y[0] = Decimal(self.__arange_y[0])
        self.__arange_y[1] = Decimal(self.__arange_y[1])
      except:
        raise ValueError("Wrong arange_y format. Must be convertible to "\
                         "Decimal.")
    elif len(self.__arange_y) != 0:
      raise ValueError("Wrong arange_y format. Must be a list"\
                      " of two elements or an empty list.")
    
    self.__density = kwargs.get("density", False)
    self.__weights = kwargs.get("weights", [])
    self.__ndecimals = kwargs.get("ndecimals", 2)
    
    self.__fixed_bin_width_x = kwargs.get("binwidth_x", -1)
    if isinstance(self.__fixed_bin_width_x, str):
      try:
        self.__fixed_bin_width_x = Decimal(self.__fixed_bin_width_x)
      except:
        raise ValueError("Can not convert", self.__fixed_bin_width_x,
                         "to decimals.")
    if self.__bins_x != -1 and self.__fixed_bin_width_x != -1:
      raise tb.ArgError("Can not specify number of bins_x "\
                              "and bin_x width at the same time.")
    if self.__bins_x == -1 and self.__fixed_bin_width_x == -1:
      raise tb.ArgError("Must specify number of bins_x "\
                              "or bin_x width.")
    
    self.__fixed_bin_width_y = kwargs.get("binwidth_y", -1)
    if isinstance(self.__fixed_bin_width_y, str):
      try:
        self.__fixed_bin_width_y = Decimal(self.__fixed_bin_width_y)
      except:
        raise ValueError("Can not convert", self.__fixed_bin_width_y,
                         "to decimals.")
    if self.__bins_y != -1 and self.__fixed_bin_width_y != -1:
      raise tb.ArgError("Can not specify number of bins_y "\
                              "and bin_y width at the same time.")
    if self.__bins_y == -1 and self.__fixed_bin_width_y == -1:
      raise tb.ArgError("Must specify number of bins_y "\
                              "or bin_y width.")
    self.__zero_to_nan = kwargs.get("zero_to_nan",False)
    self.reinitialise(True)  # Will initiate the other attributes
    self.__up_to_date = False
    self._update()
    MyHist2D.names.append(self.__name)

  def copy(self):
    """ Copy method
    
    Uses the save and read methods. Need to be able to edit /tmp
    """
    
    self.save(filename="/tmp/MyHist2D.mh")
    new_h = MyHist2D(filename="/tmp/MyHist2D.mh")
    os.system("rm /tmp/MyHist2D.mh")
    return new_h

  def read(self, filename, keep_name = False):
    """ Read the histogram from filename (string of the path to the file) """
    
    if not os.path.isfile(filename):
      raise OSError("File", filename, "not found.")
    with h5py.File(filename, "r") as ifile:
      for dsetname in ifile:
        dset = ifile[dsetname]
        if dsetname == "name":
          if not keep_name:
            self.set_name(dset[0].decode("UTF-8"), True)
        elif dsetname == "id_ndecimals":
          self.__id = int(dset[0])
          self.__ndecimals = int(dset[1])
        elif dsetname == "bins_str_x":
          self.__bins_x = dset[0].decode("UTF-8")
        elif dsetname == "bins_int_x":
          self.__bins_x = int(dset[0])
        elif dsetname == "bins_str_y":
          self.__bins_y = dset[0].decode("UTF-8")
        elif dsetname == "bins_int_y":
          self.__bins_y = int(dset[0])
        elif dsetname == "bin_width_x":
          self.__bin_width_x = float(dset[0])
        elif dsetname == "bin_width_y":
          self.__bin_width_y = float(dset[0])
        elif dsetname == "fixed_bin_width_str_x":
          self.__fixed_bin_width_x = dset[0].decode("UTF-8")
        elif dsetname == "fixed_bin_width_f_x":
          self.__fixed_bin_width_x = float(dset[0])
        elif dsetname == "fixed_bin_width_str_y":
          self.__fixed_bin_width_y = dset[0].decode("UTF-8")
        elif dsetname == "fixed_bin_width_f_y":
          self.__fixed_bin_width_y = float(dset[0])
        elif dsetname == "statistics":
          self.__mean_x = float(dset[0])
          self.__mean_y = float(dset[1])
          self.__rms_x = float(dset[2])
          self.__rms_y = float(dset[3])
          self.__integral = float(dset[4])
          self.__underflow = float(dset[5])
          self.__overflow = float(dset[6])
          self.__count = float(dset[7])
        elif dsetname == "statbox":
          self.__statbox = dset[0].decode("UTF-8")
        elif dsetname == "bools":
          self.__up_to_date = bool(dset[0])
          self.__density = bool(dset[1])
        elif dsetname == "data":
          self.__data = deepcopy(list(dset))
        elif dsetname == "arange_str_x":
          self.__arange_x = [Decimal(a.decode("UTF-8")) for a in dset]
        elif dsetname == "arange_float_x":
          self.__arange_x = [float(d) for d in list(dset)]
        elif dsetname == "arange_str_y":
          self.__arange_y = [Decimal(a.decode("UTF-8")) for a in dset]
        elif dsetname == "arange_float_y":
          self.__arange_y = [float(d) for d in list(dset)]
        elif dsetname == "weights":
          self.__weights = [float(d) for d in list(dset)]
        elif dsetname == "bin_edges_x":
          self.__bin_edges_x = [float(d) for d in list(dset)]
        elif dsetname == "bin_edges_y":
          self.__bin_edges_y = [float(d) for d in list(dset)]
        elif dsetname == "bin_middles_x":
          self.__bin_middles_x = [float(d) for d in list(dset)]
        elif dsetname == "bin_middles_y":
          self.__bin_middles_y = [float(d) for d in list(dset)]
        elif dsetname == "bin_contents":
          self.__bin_contents = deepcopy(list(dset))
        elif dsetname == "fits":
          self.__fits = [d.decode("UTF-8") for d in list(dset)]
        elif dsetname == "xyzfits":
          self.__xyzfits = deepcopy(list(dset))
        elif dsetname == "fitted_params":
          self.__fitted_params = deepcopy(list(dset))
        elif dsetname == "fitboxes":
          self.__fitboxes = [f.decode("UTF-8") for f in dset]
        elif dsetname == "chisquares":
          self.__chisquares = [float(d) for d in list(dset)]
        else:
          raise ValueError("Unkown MyHist attribute", dsetname,".")

  def load_files(self, filename, path = ""):
    """ Fill data from a simple text file
    
    Separator must be white space.
    Only uses the first two columns. Will be improved to take options like
    the same function of MyHist.
     """
    
    data = [[], []]
    with open(path+filename, "r") as ifile:
      lines = ifile.readlines()
      for line in lines:
        line = line.split("\n")[0].split(" ")
        try:
          line = [float(line[0]), float(line[1])]
          data[0].append(line[0])
          data[1].append(line[1])
        except:
          print("Line", line, "is ignored")
    self.__data = deecopy(data)
    self.reinitialise(True)
    self._update()

#Overloads
  def __str__(self, all_bins = False):
    """ Will show statbox and fitboxes
    
    And bin content alonside with bin position if all_bins is true
     """
    
    if not self.is_up_to_date():
      self._update()
    to_return = self.__statbox
    if len(self.__fitboxes) > 0:
      for f in self.__fitboxes:
        to_return = str("\n").join((to_return, f))
    if all_bins:
      for x1, ex in zip(self.__bin_contents, self.__bin_middles_x):
        for c, ey in zip(x1, self.__bin_middles_y):
          to_return = str("\n  ").join((
            to_return, str(" ").join((str(ex), str(ey), str(c)))
          ))
    return to_return

  def __getitem__(self, bin_x, bin_y):
    """get bin content based on bin number ibin."""
    
    return self.get_bin_content(bin_x, bin_y)

  def __eq__(self, other):
    """ Compares self with the instance of MyHist2D given as 'other'.

    False if other is None
    True if bin contents and bin edges are all equal
    """ 
    
    if not self.__up_to_date:
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return False
    # Should be modified to tolerate floating point imprecision
    if self.__bin_width_x != other.get_bin_width_x():
      return False
    # Should be modified to tolerate floating point imprecision
    if self.__arange_x != other.get_range_x():
      return False
    # Should be modified to tolerate floating point imprecision
    if self.__bin_width_y != other.get_bin_width_y():
      return False
    # Should be modified to tolerate floating point imprecision
    if self.__arange_y != other.get_range_y():
      return False
    if len(self.__bin_contents) != len(other.get_bin_contents()):
      return False
    if len(self.__bin_contents[0]) != len(other.get_bin_contents()[0]):
      return False
    if len(self.__bin_contents) == 0:
      return True
    for ix in range(len(self.__bin_contents)):
      xs = self.__bin_contents[xs]
      for iy in range(len(xs)):
        x1,y1 = self.get_bin_edge(ix,iy)
        x2,y2 = other.get_bin_edge(ix,iy)
        c1 = self.__bin_contents[ix][iy]
        c2 = other.__bin_contents[ix][iy]
        if c1 != c2 or x1 != x2 or y1 != y2:
          return False
    return True

  def __ne__(self,other):
    """ Compares self with the instance of MyHist given as 'other'.

    True if other is None
    False if bin contents are all equal
    and bin widths are equal and ranges are equal.
    """ 
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return True
    # Should be modified to tolerate floating point imprecision
    if self.__bin_width_x != other.get_bin_width_x():
      return True
    # Should be modified to tolerate floating point imprecision
    if self.__arange_x != other.get_range_x():
      return True
    # Should be modified to tolerate floating point imprecision
    if self.__bin_width_y != other.get_bin_width_y():
      return True
    # Should be modified to tolerate floating point imprecision
    if self.__arange_y != other.get_range_y():
      return True
    if len(self.__bin_contents) != len(other.get_bin_contents()):
      return True
    if len(self.__bin_contents[0]) != len(other.get_bin_contents()[0]):
      return True
    for ix in range(len(self.__bin_contents)):
      xs = self.__bin_contents[xs]
      for iy in range(len(xs)):
        x1,y1 = self.get_bin_edge(ix,iy)
        x2,y2 = other.get_bin_edge(ix,iy)
        c1 = self.__bin_contents[ix][iy]
        c2 = other.__bin_contents[ix][iy]
        if c1 != c2 or x1 != x2 or y1 != y2:
          return True
    return False

  def __lt__(self,other):
    """ True if self has a smaller integral than other
    
    False if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return False
    return self.__integral < other.get_integral()

  def __le__(self,other):
    """ True if self has a smaller or equal integral than other
    
    False if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return False
    return self.__integral <= other.get_integral()

  def __gt__(self,other):
    """ True if self has a bigger integral than other 
    
    True if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return True
    return self.__integral > other.get_integral()

  def __ge__(self,other):
    """ True if self has a bigger or equal integral than other
    
    True if other is None
    """
    
    if not self.is_up_to_date():
      self._update()
    if not other.is_up_to_date():
      other._update()
    if other is None:
      return True
    return self.__integral >= other.get_integral()

  def __add__(h1, h2):
    """ Addition of histograms 'h1' and 'h2'. Must be MyHist2D instances.
    
    Will create a new histogram which will be the sum of the two input histos.
    The range will cover both input histos, the bin size will be the min of
    the bin sizes of the two input histos.  Does not use the data of the two
    input histos, but directly the bin contents.
    """
    
    if not h1.is_up_to_date():
      h1._update()
    if not h2.is_up_to_date():
      h2._update()
    newrange_x = [min(h1.get_range_x()[0], h2.get_range_x()[0]),
                 max(h1.get_range_x()[1], h2.get_range_x()[1])]
    newrange_y = [min(h1.get_range_y()[0], h2.get_range_y()[0]),
                 max(h1.get_range_y()[1], h2.get_range_y()[1])]
    bin_width_x = min(h1.get_bin_width_x(), h2.get_bin_width_x())
    bin_width_y = min(h1.get_bin_width_y(), h2.get_bin_width_y())
    h = MyHist2D(name = str("").join((h1.get_name(), "+", h2.get_name())),
                 binwidth_x=bin_width_x, binwidth_y=bin_width_y)
    for x1, ex in zip(h1.get_bin_contents(),h1.get_bin_middles_x()):
      for c, ey in zip(x1, h1.get_bin_middles_y()):
       h.fill([ex, ey], c)
    for x2, ex in zip(h2.get_bin_contents(),h2.get_bin_middles_x()):
      for c, ey in zip(x2, h2.get_bin_middles_y()):
        h.fill([ex, ey], c)
    h._update()
    return h

# Histogram computing and updating methods
  def reinitialise(self, total=True):
    """ Reinitialise (or create if histogram is new) some attributes.
    
    Used any time a creation method is used, or when set_data
    method is used, or just before computing the histogram.
    Can also just reinitialize the statistical attributes (mean, rms...)
    if argument 'total' is set to False (default True).
    """
    
    if total:
      self.__bin_width_x = -1
      self.__bin_width_y = -1
      self.__bin_edges_x = []
      self.__bin_edges_y = []
      self.__bin_middles_x = []
      self.__bin_middles_y = []
      
    self.__bin_contents = [ [ 0 for j in range(len(self.__bin_middles_y)) ]
                            for i in range(len(self.__bin_middles_x)) ]
    self.__underflow = 0
    self.__overflow = 0
    self.__mean_x = self.__rms_x = self.__mean_y = self.__rms_y = 0
    self.__integral = 0
    self.__count = 0
    self.__statbox = ""
    self.__fits = []
    self.__xyzfits = []
    self.__fitted_params = []
    self.__fitboxes = []
    self.__chisquares = []
    
    self.__up_to_date = False

  def _prepare_computation(self):
    """ Determines if the histogram should be computed.
    
    See _compute_histo() for more details on the conditions.
    """
    x_ok = False
    y_ok = False
    # Histogram had already been computed.
    if len(self.__bin_edges_x) != 0 and len(self.__bin_edges_y) != 0:
      return False
    # Enough to use numpy.histogram.
    if len(self.__arange_x) != 0 and self.__bins_x != "auto":
      x_ok = True
    if len(self.__arange_y) != 0 and self.__bins_y != "auto":
      y_ok = True
    # Computes number of bins from range and specified bin width.
    if len(self.__arange_x) != 0 and self.__fixed_bin_width_x != -1:
      self._compute_number_of_bins("x")
      x_ok = True
    if len(self.__arange_y) != 0 and self.__fixed_bin_width_y != -1:
      self._compute_number_of_bins("y")
      y_ok = True

     # If range was not specified, use min and max of data. Use a bit more
     # Must ne true for boh x and y to trigger computation
    if len(self.__data[0]) != 0:
      # than the max to avoid overflow
      if len(self.__arange_x) == 0:
        self.__arange_x = [Decimal(min(self.__data[0])),
                           Decimal(max(self.__data[0])*1.0000001)]
      # Computes number of bins from range and specified bin width.
      if self.__fixed_bin_width_x != -1:
        self._compute_number_of_bins("x")
      x_ok =  True
    if len(self.__data[1]) != 0:
      if len(self.__arange_y) == 0:
        self.__arange_y = [Decimal(min(self.__data[1])),
                           Decimal(max(self.__data[1])*1.0000001)]
      if self.__fixed_bin_width_y != -1:
        self._compute_number_of_bins("y")
      y_ok =  True
    
    return x_ok and y_ok

  def _compute_histo(self):
    """ Creates the numpy.histogram object.
    
    Called by the update method only if the histogram is new.  In other words,
    only of the attributes __bin_edges[0] and __bin_edges[1] are emtpy lists.
    Will check that all attributes are compatible with one another and are ok.
    Computes the bin contents, edges and middles.
    Computes the number of overflow and underflow in the data set.
    If bins was set to auto, it will then be updated with the resulting number
    of bins.
    If range was set in decimal format, it will now be floating values.
    Data and weights are emptied upon computation.
    
    The histogram can be created completely empty.  In that case, nothing is
    to be done.  Empty means that there is no way to establish the range,
    number of bins and bin width of the x or y axis.
    So, the histogram is computed if:
      - The data set is not empty.  In that case:
        - If range was not given, it is assumed to be the min and max of the
          data set.
        - If bins was given as argument, it is used.
        - If binwidth was given as argument, computes bins from it and range.
        - Those cases do not have to be the same for x and y
      - Having an empty data set but specifying either bins and range or
        binwidth and range.
    """
    
    self.reinitialise(True)
    if not self._prepare_computation():
      return
    
    if len(self.__weights) != 0:
      if self.__bins_x == "auto" or self.__bins_y == "auto":
        raise tb.ArgError("Cannot have auto bins and weighted bins at "\
                               "the same time for x axis.")
      if (len(self.__data[0]) != len(self.__weights)
        or len(self.__data[1]) != len(self.__weights)):
        raise tb.ArgError("__weights and __data must have same size.")
        
    # Check if range makes sense.
    if self.__arange_x[0] > self.__arange_x[1]:
      tmp = self.__arange_x[0]
      self.__arange_x[0] = self.__arange_x[1]
      self.__arange_x[1] = tmp
    if self.__arange_y[0] > self.__arange_y[1]:
      tmp = self.__arange_y[0]
      self.__arange_y[0] = self.__arange_y[1]
      self.__arange_y[1] = tmp
    # Convert decimal range to float.
    tmp_range_x = [float(self.__arange_x[0]), float(self.__arange_x[1])]
    tmp_range_y = [float(self.__arange_y[0]), float(self.__arange_y[1])]
    # Compute histo even if data is empty (bin edges will be up to date).
    if len(self.__weights) == 0:
      self.__weights = None
    self.__bin_contents, self.__bin_edges_x, self.__bin_edges_y = np.histogram2d(
       self.__data[0], self.__data[1],
       bins = [self.__bins_x, self.__bins_y],
       range = [tmp_range_x, tmp_range_y],
       weights = self.__weights
     )
    self.__bin_contents = list(self.__bin_contents)
    self.__bin_edges_x = list(self.__bin_edges_x)
    self.__bin_edges_y = list(self.__bin_edges_y)
    self.__bin_width_x = self.__bin_edges_x[1] - self.__bin_edges_x[0]
    self.__bin_width_y = self.__bin_edges_y[1] - self.__bin_edges_y[0]
    if self.__bins_x == "auto":
      self.__bins_x = len(self.__bin_contents[0])
      self.__bins_y = len(self.__bin_contents[0][0])
    for x, y in zip(self.__data[0], self.__data[1]):
      if x < self.__bin_edges_x[0] or y < self.__bin_edges_y[0]:
        self.__underflow += 1
      if x > self.__bin_edges_x[-1] or y > self.__bin_edges_y[-1]:
        self.__overflow += 1
    # Compute center positions of bins.and empty data and weights.
    self.__bin_middles_x = [ (self.__bin_edges_x[i-1]+self.__bin_edges_x[i])/2
                         for i in range(1, len(self.__bin_edges_x)) ]
    self.__bin_middles_y = [ (self.__bin_edges_y[i-1]+self.__bin_edges_y[i])/2
                         for i in range(1, len(self.__bin_edges_y)) ]
    self.__data = [[],[]]
    self.__weights = []

  def _update(self):
    """ Update the statistical attributes of the object.
    
    Execute _compute_histo() if histogram is empty. Computing histogram will
    reinitialise all statistical and histogram attributes.
    If not empty, will only update the statistical attributes.
    
    Runs at object creation and before any kind of display, get or fit.
    User should not have to use it as it is supposed to run when needed.
    """
    
    if len(self.__bin_edges_x) == 0 and len(self.__bin_edges_y) == 0:
      self._compute_histo()
    self.__up_to_date = True
    if len(self.__bin_edges_x) == 0 and  len(self.__bin_edges_y) == 0:
      return
    
    # Computes the statistical attributes.
    self.__count = 0
    self.__mean_x = 0.
    self.__mean_y = 0.
    sqmean_x = 0.
    sqmean_y = 0.
    self.__rms_x = 0.
    self.__rms_y = 0.
    self.__integral = 0
    ybin_content = {}
    for x, mx in zip(self.__bin_contents, self.__bin_middles_x):
      xbin_content = 0
      for c, my in zip(x,self.__bin_middles_y):
        if not c == c:  # Empty bins can be np.nan
          continue
        if my in ybin_content:
          ybin_content[my] += c
        else:
          ybin_content[my] = c
        xbin_content += c
        self.__count = self.__count + c  # Can be float.
        self.__integral += c*self.__bin_width_x*self.__bin_width_y
      self.__mean_x = self.__mean_x + xbin_content*mx
      sqmean_x = sqmean_x + xbin_content*mx**2
    self.__mean_y = sum([y*ybin_content[y] for y in ybin_content])
    sqmean_y = sum([(y**2)*ybin_content[y] for y in ybin_content])
    if self.__count > 0:
      self.__mean_x = self.__mean_x/self.__count
      sqmean_x = sqmean_x/self.__count
      self.__mean_y = self.__mean_y/self.__count
      sqmean_y = sqmean_y/self.__count
    try:
      self.__rms_x = math.sqrt(sqmean_x - self.__mean_x**2)
    except ValueError:
      self.__rms_x = 0.
    try:
      self.__rms_y = math.sqrt(sqmean_y - self.__mean_y**2)
    except ValueError:
      self.__rms_y = 0.
    if self.__density and self.__integral > 0:
      for ix in range(len(self.__bin_contents)):
        for iy in range(len(self.__bin_contents[0])):
          if self.__bin_contents[ix][iy] == self.__bin_contents[ix][iy]:
            self.__bin_contents[ix][iy] = (self.__bin_contents[ix][iy]
                                         / self.__integral)
      self.__integral = 0
      for x, mx in zip(self.__bin_contents, self.__bin_middles_x):
        xbin_content = 0
        for c, my in zip(x,self.__bin_middles_y):
          if not c == c:  # Empty bins can be np.nan
            continue
          self.__integral += c*self.__bin_width_x*self.__bin_width_y
    
    if self.__zero_to_nan:
      for ix in range(len(self.__bin_contents)):
        for iy in range(len(self.__bin_contents[0])):
          if self.__bin_contents[ix][iy] == 0:
            self.__bin_contents[ix][iy] = np.nan
    self._form_stat_box()

  def _compute_number_of_bins(self, axis):
    """ Compute number of bins based on range and fixed bin size.
    
    Used by _prepare_computation().
    """
    
    if axis == "x":
      if self.__arange_x[0] > self.__arange_x[1]:
        self.__arange_x = [self.__arange_x[1], self.__arange_x[0]]
      self.__bins_x = round(
                        (self.__arange_x[1]-self.__arange_x[0])
                        / Decimal(self.__fixed_bin_width_x)
                       )
      if self.__bins_x <= 0:
        raise ValueError("Number of bins must be positive.")
    if axis == "y":
      if self.__arange_y[0] > self.__arange_y[1]:
        self.__arange_y = [self.__arange_y[1], self.__arange_y[0]]
      self.__bins_y = round(
                        (self.__arange_y[1]-self.__arange_y[0])
                        / Decimal(self.__fixed_bin_width_y)
                       )
      if self.__bins_y <= 0:
        raise ValueError("Number of bins must be positive.")
    self.__up_to_date = False


# Set and add methods.
  def set_name(self, name, iterate = False):
    """ Sets the name of the instance to the given string. """
    
    if name in MyHist2D.names:
      if iterate:
        iname = 2
        name_tmp = str("_").join((name, str(iname)))
        while name_tmp in MyHist2D.names:
          iname += 1
          name_tmp = str("_").join((name, str(iname)))
        name = name_tmp
      else:
        warnings.warn(str(" ").join(("Name", name, "is already used.")),
                      Warning)
    self.__name = name
    MyHist2D.names.append(self.__name)

  def set_data(self, data, reinitialise = True):
    """ Set the data attribute and recomputes the histogram if asked.
    
    Arguments are:
      data -- the list to set __data to
      reinitialise -- If True, will reinitialise the histogram. (Default True)
    """ 
    
    try:
      if len(data[0]) != len(data[1]):
        raise ValueError("data format incorrect.")
    except:
        raise ValueError("data format incorrect.")
    self.__data = data[:]
    if reinitialise:
      self.reinitialise(True)
      self._update()
    else:
      self.__up_to_date = False

  def set_number_of_bins_x(self, bins):
    """ Set the 'bins' attribute to the integer or string given as argument.
    
    Only makes sense to do that is bin contents and edges have not been
    computed yet.  Will implement a rebin method for that purpose.
    Use 'auto' to reset the value.
    """
    
    if len(self.__bin_edges_x) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if self.__fixed_bin_width_x != -1 and bins != "auto":
      raise tb.ArgError("can not set __bins if __fixed_bin_width was "\
                             "specified.")
    self.__bins_x = int(bins)
    self.__up_to_date = False

  def set_number_of_bins_y(self, bins):
    """ Set the 'bins' attribute to the integer or string given as argument.
    
    Only makes sense to do that is bin contents and edges have not been
    computed yet.  Will implement a rebin method for that purpose.
    Use 'auto' to reset the value.
    """
    
    if len(self.__bin_edges_y) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if self.__fixed_bin_width_y != -1 and bins != "auto":
      raise tb.ArgError("can not set __bins if __fixed_bin_width was "\
                             "specified.")
    self.__bins_y = int(bins)
    self.__up_to_date = False

  def set_range_x(self, arange):
    """ Set the 'arange' attribute.
    
    Argument must be a list of min and max range.  Can be string, if so will
    convert it to decimal for better precision.
    Only makes sense to do that is bin contents and edges have not been
    computed yet.  Will implement a specific method for that purpose.
    Use [] to reset the value.
    """
    
    if len(self.__bin_edges_x) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if len(arange) != 2 and len(arange) != 0:
      raise ValueError("Wrong format. arange must be list of two elements or"\
                       " an empty list.")
    if ((not type(arange[0]) is Decimal and
         not type(arange[0]) is int and
         not type(arange[0]) is str)):
      warnings.warn("You should use Decimal instead of a plain float for "\
                    "arange to avoid floating point imprecisions.", Warning)
    arange[0] = Decimal(arange[0])
    arange[1] = Decimal(arange[1])
    self.__arange_x = [min(arange), max(arange)]
    self.__up_to_date = False

  def set_range_y(self, arange):
    """ Set the 'arange' attribute.
    
    Argument must be a list of min and max range.  Can be string, if so will
    convert it to decimal for better precision.
    Only makes sense to do that is bin contents and edges have not been
    computed yet.  Will implement a specific method for that purpose.
    Use [] to reset the value.
    """
    
    if len(self.__bin_edges_y) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if len(arange) != 2 and len(arange) != 0:
      raise ValueError("Wrong format. arange must be list of two elements or"\
                       " an empty list.")
    if ((not type(arange[0]) is Decimal and
         not type(arange[0]) is int and
         not type(arange[0]) is str)):
      warnings.warn("You should use Decimal instead of a plain float for "\
                    "arange to avoid floating point imprecisions.", Warning)
    arange[0] = Decimal(arange[0])
    arange[1] = Decimal(arange[1])
    self.__arange_y = [min(arange), max(arange)]
    self.__up_to_date = False

  def set_weights(self, weights, reinitialise=True):
    """ Set the weights attribute and recomputes the histogram if asked.
    
    Arguments are:
      weights -- list of weights to set
      reinitialise -- If True, will reinitialise the histogram. (Default True)
    """
    
    self.__weights = weights
    if reinitialise:
      self.reinitialise(True)
      self._update()
    else:
      self.__up_to_date = False

  def set_bin_width_x(self, binwidth):
    """ Set the 'fixed_bins_size' attribute to the given argument.
    
    Can be string, if so will convert it to decimal for better precision.
    Only makes sense to do that if bin contents and edges have not been
    computed yet.  Will implement a rebin method for that purpose.
    Use -1 to reset the value.
    """
    
    if len(self.__bin_edges_x) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if self.__bins_x != "auto" and binwidth != -1:
      raise tb.ArgError("Can not set binwidth if __bins "\
                             "was specified.")
    if isinstance(binwidth, str):
      try:
        binwidth = Decimal(binwidth)
      except:
        raise ValueError("Can not convert", binwidth, "to decimals.")
    if not isinstance(binwidth, Decimal) and not isinstance(binwidth, int):
      warnings.warn("You should use Decimal instead of a plain float for "\
                    "binwidth to avoid floating point imprecisions.", Warning)
    self.__fixed_bin_width_x = binwidth
    self.__up_to_date = False

  def set_bin_width_y(self, binwidth):
    """ Set the 'fixed_bins_size' attribute to the given argument.
    
    Can be string, if so will convert it to decimal for better precision.
    Only makes sense to do that if bin contents and edges have not been
    computed yet.  Will implement a rebin method for that purpose.
    Use -1 to reset the value.
    """
    
    if len(self.__bin_edges_y) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    if self.__bins_y != "auto" and binwidth != -1:
      raise tb.ArgError("Can not set binwidth if __bins "\
                             "was specified.")
    if isinstance(binwidth, str):
      try:
        binwidth = Decimal(binwidth)
      except:
        raise ValueError("Can not convert", binwidth, "to decimals.")
    if not isinstance(binwidth, Decimal) and not isinstance(binwidth, int):
      warnings.warn("You should use Decimal instead of a plain float for "\
                    "binwidth to avoid floating point imprecisions.", Warning)
    self.__fixed_bin_width_y = binwidth
    self.__up_to_date = False

  def set_up_to_date(self):
    """ Tells the object it is not up to date.
    
    Not sure when this should be used. Probably never.
    """
    self.__up_to_date = True

  def fill(self, values, weights = None):
    """Add xs and ys values with weights to the appropriate bin.
    
    Arguments are:
      values -- One or a list of pairs [x,y]
      weights -- weights to add on each x. Can be number-like or
        iterable-like.
    """
    if not isinstance(values, collections.Iterable):  # values must be iterable
      raise ValueError("Incorrect format.")
    if len(values) == 0:  # If values is empty, nothing to do
      return
    if len(values) == 2:  # If values is [x,y], transform to [[x,y]]
      if not isinstance(values[0], collections.Iterable):
        if (isinstance(values[0], numbers.Number)
          and isinstance(values[1], numbers.Number)):  # skip np.nan
          values = [values]
        else:
          return
    
    update_weights = True
    if not isinstance(weights, collections.Iterable):
      if weights is None:
        update_weights = False
        weights = [1]*len(values)
      elif isinstance(weights, numbers.Number):
        weights = [weights]
      else:
        return  # skip np.nan
    
    if len(values) != len(weights):
      raise tb.ArgError("values and weights must have the same length.")
    # If histogram is new. In that case, append values to data and weights
    # to weights.
    if len(self.__bin_edges_x) == 0 and len(self.__bin_edges_x) == 0:
      for value, weight in zip(values, weights):
        if not isinstance(weight, numbers.Number):
          continue
        if (not isinstance(value[0], numbers.Number)
          or not isinstance(value[1], numbers.Number)):
          continue
        self.__data[0].append(value[0])
        self.__data[1].append(value[1])
        if update_weights:
          self.__weights.append(weight)
      self.__up_to_date = False
      return

    for v, w in zip(values, weights):
      ibin_x = self.get_bin_x(v[0])
      ibin_y = self.get_bin_y(v[1])
      if ibin_x == "underflow" or ibin_y == "underflow":
        self.__underflow += w
      elif ibin_x == "overflow" or ibin_y == "overflow":
        self.__overflow += w
      else:
        if len(self.__fits) != 0:
          self.__fits = []# If bin content is modified, the current fits 
                          # are meaningless and are suppressed.
        self.add_bin_content(ibin_x, ibin_y, w)

  def add_bin_content(self, ibin_x, ibin_y, weight = 1):
    """Add a weight (can be float) to bin_x-bin_y.
    
    Histogram needs to have been computed first.
    """
    
    if len(self.__bin_edges_x) == 0 and len(self.__data[0]) != 0:
      self._update()
    elif len(self.__bin_edges_x) == 0:
      raise ValueError("Histogram is empty.")
    if (not self.__bin_contents[ibin_x][ibin_y]
        == self.__bin_contents[ibin_x][ibin_y]):  # is np.nan
      self.set_bin_content(ibin_x, ibin_y, weight)
    else:
      self.__bin_contents[ibin_x][ibin_y] += weight
    self.__up_to_date = False

  def set_bin_content(self, ibin_x, ibin_y, value):
    """Set a value to bin ibin.
    
    Histogram needs to have been computed first.
    """
    
    if len(self.__bin_edges_x) == 0 and len(self.__data[0]) != 0:
      self._update()
    elif len(self.__bin_edges_x) == 0:
      raise ValueError("Histogram is empty.")
    if (ibin_x == "overflow" or ibin_x == "underflow" or
       ibin_y == "overflow" or ibin_y == "underflow"):
      raise ValueError("Bin out of range")
    if (ibin_x < 0 or ibin_x > len(self.__bin_contents) or
        ibin_y < 0 or ibin_y > len(self.__bin_contents[0])):
      raise ValueError("Bin out of range")
    self.__bin_contents[ibin_x][ibin_y] = value
    self.__up_to_date = False

  def set_normalise(self, density = True):
    """ Set the 'density' attribute to the given argument (Default True).
    
    Only makes sense to do that is histogram has not been computed yet.
    """
    
    if len(self.__bin_edges_x) != 0:
      raise tb.ArgError("Can not be done on a histogram "\
                             "that has already been computed.")
    self.__density = bool(density)
    self.__up_to_date = False

  def set_decimals(self, value):
    """ Set the number of decimals to be used to display histogram stats.
    
    Will update statbox and fitbox.
    """
    
    self.__ndecimals = int(value)
    if self.__statbox != "":
      self._form_stat_box()
    if len(self.__fitboxes) != 0:
      self._form_fit_box()
    

#get methods
  def get_id(self):
    return self.__id

  def get_name(self):
    return self.__name

  def get_label(self):
    return self.__name

  def get_data(self):
    return deepcopy(self.__data)

  def get_number_of_bins_x(self, do_not_update = False):
    if not self.__up_to_date and not do_not_update:
      self._update()
    return self.__bins_x

  def get_number_of_bins_y(self, do_not_update = False):
    if not self.__up_to_date and not do_not_update:
      self._update()
    return self.__bins_y

  def get_range_x(self):
    if not self.__up_to_date:
      self._update()
    return self.__arange_x[:]

  def get_drange_x(self):
    if not self.__up_to_date:
      self._update()
    return abs(self.__arange_x[1]-self.__arange_x[0])

  def get_range_y(self):
    if not self.__up_to_date:
      self._update()
    return self.__arange_y[:]

  def get_drange_y(self):
    if not self.__up_to_date:
      self._update()
    return abs(self.__arange_y[1]-self.__arange_y[0])

  def is_normalised(self):
    return self.__density

  def get_weights(self):
    return self.__weights[:]

  def get_ndecimals(self):
    return self.__ndecimals

  def get_fixed_bin_width_x(self):
    return self.__fixed_bin_width_x

  def get_fixed_bin_width_y(self):
    return self.__fixed_bin_width_y

  def is_up_to_date(self):
    return self.__up_to_date

  def get_bin_width_x(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_width_x

  def get_bin_width_y(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_width_y

  def get_bin_edges_x(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_edges_x[:]

  def get_bin_edges_y(self):
    if not self.is_up_to_date():
      self._update()
    return self.__bin_edges_y[:]

  def get_bin_middles_x(self, do_not_update = False):
    if not self.is_up_to_date() and not do_not_update:
      self._update()
    return self.__bin_middles_x[:]

  def get_bin_middles_y(self, do_not_update = False):
    if not self.is_up_to_date() and not do_not_update:
      self._update()
    return self.__bin_middles_y[:]

  def get_bin_contents(self):
    if not self.is_up_to_date():
      self._update()
    return deepcopy(self.__bin_contents)
    
  def get_underflow(self):
    if not self.is_up_to_date():
      self._update()
    return self.__underflow
    
  def get_overflow(self):
    if not self.is_up_to_date():
      self._update()
    return self.__overflow

  def get_mean_x(self):
    if not self.is_up_to_date():
      self._update()
    return self.__mean_x

  def get_mean_y(self):
    if not self.is_up_to_date():
      self._update()
    return self.__mean_y
    
  def get_std_dev_x(self):
    if not self.is_up_to_date():
      self._update()
    return self.__rms_x
    
  def get_std_dev_y(self):
    if not self.is_up_to_date():
      self._update()
    return self.__rms_y
    
  def get_integral(self):
    if not self.is_up_to_date():
      self._update()
    return self.__integral
    
  def get_entries(self):
    if not self.is_up_to_date():
      self._update()
    return self.__count

  def get_statbox(self):
    if not self.is_up_to_date():
      self._update()
    return self.__statbox[:]

  def get_fits(self):
    return self.__fits[:]

  def get_xyzfits(self):
    return self.__xyzfits[:]

  def get_fitted_params(self):
    return self.__fitted_params[:]

  def get_fitboxes(self):
    return self.__fitboxes[:]

  def get_chisquares(self):
    return self.__chisquares[:]

  def get_bin_x(self, value):
    """ Returns the bin number of the given value.
    
    Histogram needs to have been computed first, but does not need to be
    up to date since the content of the bin if not used.
    """
    
    if (len(self.__bin_edges_x) == 0 and len(self.__data[0]) != 0):
      self._update()
    elif len(self.__bin_edges_x) == 0:
      raise ValueError("Histogram is empty.")
    if value >= self.__bin_edges_x[-1]:
      return "overflow"
    if value < self.__bin_edges_x[0]:
      return "underflow"
    return int((Decimal(value)-self.__arange_x[0])/Decimal(self.__bin_width_x))

  def get_bin_y(self, value):
    if (len(self.__bin_edges_y) == 0 and len(self.__data[0]) != 0):
      self._update()
    elif len(self.__bin_edges_y) == 0:
      raise ValueError("Histogram is empty.")
    if value >= self.__bin_edges_y[-1]:
      return "overflow"
    if value < self.__bin_edges_y[0]:
      return "underflow"
    return int((Decimal(value)-self.__arange_y[0])/Decimal(self.__bin_width_y))

  def get_bin_content(self, ibin_x, ibin_y):
    """ Returns the content of bin_x-bin_y.
    
    Returns None if ibins are not in range of bin_contents.
    """
    
    if not self.is_up_to_date():
      self._update()
    if ibin_x < 0 or ibin_x > len(self.__bin_contents):
      return None
    if ibin_y < 0 or ibin_y > len(self.__bin_contents[0]):
      return None
    return self.__bin_contents[ibin_x][ibin_y]

  def get_bin_edges(self, ibin_x, ibin_y):
    """ Returns the [left, bottom] edges of the bin.
    
    Returns None if ibin_x and ibin_y are not in range of histo
    """
    
    if not self.is_up_to_date():
      self._update()
    if ibin_x < 0 or ibin_x > len(self.__bin_contents):
      return None
    if ibin_y < 0 or ibin_y > len(self.__bin_contents[0]):
      return None
    return [self.__bin_edges_x[ibin_x], self.__bin_edges_y[ibin_y]]

  def get_bin_middle_x(self, ibin_x, do_not_update = False):
    """ Returns the middle_x of the bin.
    
    Returns None if ibin_x is not in range of histo
    """
    
    if not self.is_up_to_date() and not do_not_update:
      self._update()
    if ibin_x < 0 or ibin_x > len(self.__bin_contents):
      return None
    return self.__bin_middles_x[ibin_x]

  def get_bin_middle_y(self, ibin_y, do_not_update = False):
    """ Returns the middle_y of the bin.
    
    Returns None if ibin_y is not in range of histo
    """
    
    if not self.is_up_to_date() and not do_not_update:
      self._update()
    if ibin_y < 0 or ibin_y > len(self.__bin_contents[0]):
      return None
    return self.__bin_middles_y[ibin_y]

  def get_bin_middles(self, ibin_x, ibin_y):
    """ Returns the [middle_x, middle_y] of the bin.
    
    Returns None if ibin_x and ibin_y are not in range of histo
    """
    
    if not self.is_up_to_date():
      self._update()
    if ibin_x < 0 or ibin_x > len(self.__bin_contents):
      return None
    if ibin_y < 0 or ibin_y > len(self.__bin_contents[0]):
      return None
    return [self.__bin_middles_x[ibin_x], self.__bin_middles_y[ibin_y]]

  def get_xy_content(self, x, y):
    """ Returns the content of the bin corresponding to the given x and y.
    
    Returns None if x and y are not in range of histo
    """
    
    if not self.is_up_to_date():
      self._update()
    ibin_x = self.get_bin_x(x)
    ibin_y = self.get_bin_y(x)
    return self.get_bin_content(ibin_x, ibin_y)

  def get_maximum(self):
    """ Returns the maximum count.
    
    Histogram needs to have been computed first and up to date.
    """
    
    if not self.is_up_to_date():
      self._update()
    elif len(self.__bin_edges_x) == 0:
      raise ValueError("Histogram is empty.")
    return max([max(l) for l in self.__bin_contents])

  def get_position_of_maximum(self):
    """ Returns the position of the maximum count.
    
    Histogram needs to have been computed first and up to date.
    """
    
    maximum = self.get_maximum()
    for x, mx in zip(self.__bin_contents, self.__bin_middles_x):
      for c, my in zip(x, self.__bin_middles_y):
        if c >= maximum:
          return mx, my
      maximum = maximum*0.9999
      for c, my in zip(x, self.__bin_middles_y):
        if c >= maximum:
          return mx, my


# Display methods.
  def _form_stat_box(self):
    if not self.is_up_to_date():
      self._update()
    if self.__count is None:
      return None
    Entries = self.__count
    if Entries != int(Entries):
      Entries = tb.stround(Entries, self.__ndecimals)
    else:
      Entries = str(int(Entries))
    to_return = str("").join((self.__name , ":"))
    to_return = str("\n ").join((
                   to_return,
                   str(" ").join(("count:", Entries)),
                   str(" ").join(("mean x:", tb.stround(self.__mean_x,
                                                      self.__ndecimals))),
                   str(" ").join(("mean y:", tb.stround(self.__mean_y,
                                                      self.__ndecimals))),
                   str(" ").join(("Std Dev x:", tb.stround(
                                                      self.__rms_x,
                                                      self.__ndecimals))),
                   str(" ").join(("Std Dev y:", tb.stround(
                                                      self.__rms_y,
                                                      self.__ndecimals))),
                   str(" ").join(("Integral:", tb.stround(
                                                      self.__integral,
                                                      self.__ndecimals)))
                 ))
    if self.__underflow > 0:
      to_return = str("\n ").join((
                     to_return,
                     str(" ").join(("Underflow:", str(self.__underflow)))
                   ))
    if self.__overflow > 0:
      to_return = str("\n ").join((
                     to_return,
                     str(" ").join(("Overflow:", str(self.__overflow)))
                   ))
    self.__statbox = to_return

  def _form_fit_box(self):
    if len(self.__fits) == 0:
      return
    if len(self.__fitboxes) == 0:
      self.__fitboxes.append(str("").join((self.__name, ":")))
    fitstring = str("\n").join((" Fitted By:",self.__fits[-1]))
    for ipar in range(len(self.__fitted_params[-1])):
      fitstring = str("\n  ").join((
            fitstring,
            str("").join(("p[",str(ipar),"]: ",
              tb.stround(self.__fitted_params[-1][ipar], 
              self.__ndecimals)
            ))
          ))
    fitstring = str("\n  ").join((
            fitstring,
            str("=").join(("$\chi^2$",
              tb.stround(self.__chisquares[-1],
              self.__ndecimals)
            ))
          ))
    self.__fitboxes[-1] = str("").join((self.__fitboxes[-1], fitstring))

  def save(self, filename="", path = ""):
    """ Save the histogram to a file, readable with read method """
    
    if path != "":
      if not os.path.isdir(path):
        raise OSError("Directory",path,"not found.")
    if not self.is_up_to_date():
      self._update()
    if filename == "":
      filename = str("").join(("MyHist2D_", self.__name, ".hdf5"))
    fullname = str("").join((path, filename))
    if "/tmp" not in fullname:
      print(MyHist2D.skip, "Info in MyHist2D::save: Saving histogram in file",
            fullname)
    with h5py.File(fullname,"w") as ofile:
      ofile.create_dataset("name", data=[np.string_(self.__name)])
      ofile.create_dataset("id_ndecimals",
                            data=[self.__id, self.__ndecimals], dtype='i8')
      if isinstance(self.__bins_x, str):
        ofile.create_dataset("bins_str_x", data=[np.string_(self.__bins_x)])
      else:
        ofile.create_dataset("bins_int_x", data=[self.__bins_x], dtype='i8')
      if isinstance(self.__bins_y, str):
        ofile.create_dataset("bins_str_y", data=[np.string_(self.__bins_y)])
      else:
        ofile.create_dataset("bins_int_y", data=[self.__bins_y], dtype='i8')
      ofile.create_dataset("bin_width_x",
           data=[self.__bin_width_x], dtype='f8')
      ofile.create_dataset("bin_width_y",
           data=[self.__bin_width_y], dtype='f8')
      if isinstance(self.__fixed_bin_width_x, Decimal):
        ofile.create_dataset("fixed_bin_width_str_x",
                              data=[np.string_(self.__fixed_bin_width_x)])
      else:
        ofile.create_dataset("fixed_bin_width_f_x",
                             data=[self.__fixed_bin_width_x], dtype='f8')
      if isinstance(self.__fixed_bin_width_y, Decimal):
        ofile.create_dataset("fixed_bin_width_str_y",
                              data=[np.string_(self.__fixed_bin_width_y)])
      else:
        ofile.create_dataset("fixed_bin_width_f_y",
                             data=[self.__fixed_bin_width_y], dtype='f8')
      ofile.create_dataset("statistics", data=[self.__mean_x,
                                              self.__mean_y,
                                              self.__rms_x,
                                              self.__rms_y,
                                              self.__integral,
                                              self.__underflow,
                                              self.__overflow,
                                              self.__count], dtype='f8')
      ofile.create_dataset("statbox",
                           data=[np.string_(self.__statbox)])
      ofile.create_dataset("bools", data=[self.__up_to_date,
                                          self.__density], dtype='?')
      ofile.create_dataset("data", data=deepcopy(self.__data), dtype='f8')
      if isinstance(self.__arange_x[0], Decimal):
        ofile.create_dataset("arange_str_x", data=[np.string_(a)
                                             for a in self.__arange_x])
      else:
        ofile.create_dataset("arange_float_x",
           data=self.__arange_x[:], dtype='f8')
      if isinstance(self.__arange_y[0], Decimal):
        ofile.create_dataset("arange_str_y", data=[np.string_(a)
                                             for a in self.__arange_y])
      else:
        ofile.create_dataset("arange_float_y",
           data=self.__arange_y[:], dtype='f8')
      ofile.create_dataset("weights", data=self.__weights[:], dtype='f8')
      ofile.create_dataset("bin_edges_x",
           data=self.__bin_edges_x[:], dtype='f8')
      ofile.create_dataset("bin_edges_y",
           data=self.__bin_edges_y[:], dtype='f8')
      ofile.create_dataset("bin_middles_x",
           data=self.__bin_middles_x[:], dtype='f8')
      ofile.create_dataset("bin_middles_y",
           data=self.__bin_middles_y[:], dtype='f8')
      ofile.create_dataset("bin_contents",
           data=deepcopy(self.__bin_contents), dtype='f8')
      ofile.create_dataset("fits", data=[np.string_(f) for f in self.__fits])
      ofile.create_dataset("xyzfits", data=deepcopy(self.__xyzfits), dtype='f8')
      ofile.create_dataset("fitted_params",
           data=self.__fitted_params[:], dtype='f8')
      ofile.create_dataset("fitboxes", data=[np.string_(f)
                                             for f in self.__fitboxes])
      ofile.create_dataset("chisquares",
           data=self.__chisquares[:], dtype='f8')

  def form_subplots_for_clb_and_leg(self, hide_statbox, hide_fitbox, hide_clb,
                                    aspect, figsize):
    fig = None
    gs = None
    ax1 = None
    ax_leg = None
    ax_clb = None
    wspace=0.1
    main_figsize = figsize
    if main_figsize is None:
      main_figsize = 10
    if not hide_clb:
      if aspect <= 1.5:
        if ((len(self.__fitboxes) != 0 and not hide_fitbox)
          or (self.__statbox != "" and not hide_statbox)):
          widths = [main_figsize-3*wspace, main_figsize/20, main_figsize/3]
          heights = [main_figsize*aspect]
          fig = plt.figure(figsize=(sum(widths),heights[0]))
          
          gs = gridspec.GridSpec(ncols=3, nrows=1,
                                 width_ratios=widths,
                                 height_ratios=heights)
          ax1 = fig.add_subplot(gs[0, 0])
          ax_clb = fig.add_subplot(gs[0, 1])
          ax_leg = fig.add_subplot(gs[0, 2])
          ax_leg.axis('off')
          plt.subplots_adjust(wspace=wspace)
        else:
          widths = [main_figsize, main_figsize/20]
          heights = [main_figsize*aspect]
          fig = plt.figure(figsize=(sum(widths),heights[0]))
          
          gs = gridspec.GridSpec(ncols=2, nrows=1,
                                 width_ratios=widths,
                                 height_ratios=heights)
          ax1 = fig.add_subplot(gs[0, 0])
          ax_clb = fig.add_subplot(gs[0, 1])
          plt.subplots_adjust(wspace=wspace)
      else:
        if ((len(self.__fitboxes) != 0 and not hide_fitbox)
          or (self.__statbox != "" and not hide_statbox)):
          fig = plt.figure(
                           constrained_layout=False,
                           figsize = (main_figsize+main_figsize/4,
                                      main_figsize*aspect+11/10)
                          )
          gs = gridspec.GridSpec(2, 2,
                                 width_ratios=[main_figsize/4, 1],
                                 height_ratios=[10, 1])

          ax1 = plt.subplot(gs[0,0])
          ax_leg = plt.subplot(gs[:,1])
          ax_leg.axis('off')
          ax_clb = plt.subplot(gs[1,0])
        else:
          fig = plt.figure(
                           constrained_layout=False,
                           figsize = (main_figsize,
                                      main_figsize*aspect+11/10)
                          )
          gs = gridspec.GridSpec(ncols=1, nrows=2,
                                 width_ratios=[main_figsize],
                                 height_ratios=[10, 1])

          ax1 = plt.subplot(gs[0,0])
          ax_clb = plt.subplot(gs[1,0])
    else:
      if ((len(self.__fitboxes) != 0 and not hide_fitbox)
          or (self.__statbox != "" and not hide_statbox)):
        widths = [main_figsize, main_figsize/4]
        heights = [main_figsize*aspect]
        fig = plt.figure(figsize=(sum(widths), heights[0]))
        
        gs = gridspec.GridSpec(ncols=2, nrows=1,
                               width_ratios=widths,
                               height_ratios=heights)
        ax1 = fig.add_subplot(gs[0, 0])
        ax_leg = fig.add_subplot(gs[0, 1])
        ax_leg.axis('off')
        plt.subplots_adjust(wspace=wspace)
      else:
        fig = plt.figure(figsize=(main_figsize, main_figsize*aspect))
        ax1 = plt.subplot(111)
    return fig, ax1, ax_clb, ax_leg

  def plot(self, **kwargs):
    """ Method to plot and/or save the histogram using imshow.
    
    z axis means colormap axis.
    All arguments are optional.
    Arguments are:
      (Axes) ax -- ax object to plot in (default None)
      (pltpdf) pdf -- will save the plot as a new page in the given pdf.
        Can not be specified along with the ax argument. (default None)
      (plt.cm) cm -- matplotlib colormap to use (default plt.cm.plasma)
      (strings) xlabel, ylabel, zlabel, title -- self explanatory (default None).
      (string) option -- if contains 'save', will save the plot in a pdf file
        named 'MyHist_self.__name.pdf' (independently from the pdf argument).
        If it contains 'show', will show the plot on screen.  Both can be used
        at the same time, either by writing 'save show', 'show save', or
        without spaces. (default '')
      (string) fitcolor -- color to use to plot fitted function. (default None)
      -(list) x/y/zticks -- list of ticks to use on x/y/z axis. (default None)
      -(string) x/y/zticksformat: used to specify format of x/y/z tick labels.
        (default None)
      -(float) x/y/zticksfontsize -- self explanatory
      -(bool) force_x/ylim -- will force the range of x/y axis to match the
        range of the histogram, +/- 1%
      -(bool) adapt_x/ylim -- if ax is specified and the current limits
        do not allow to display the entire histo, will change them
        so they do (default False).
      -(bool) hide_statbox/fitbox/clb -- self explanatory (clb=colorbar)
      -(float) aspect -- y over x axis size ratio (default ratio of ranges
        size)
      -(float) x/y/zticksspacing -- self explanatory
      -(list) x/y/ztickslabels -- self explanatory
      -(list) x/y/zlabelfontsize -- self explanatory
      -(float) figsize -- size of the figure
      -(bool) blank_zero -- will set bins with 0 entries to np.nan to print
                            blank bins
    
    If no pdf or ax is specified, will return a new plt.subplot ax containing
    the plot of the histogram, with all its fitted functions.
    """
    
    expected = ["ax", "pdf", "cm", "xlabel", "ylabel", "zlabel",
                "xlabelfontsize", "ylabelfontsize", "zlabelfontsize",
                "title", "titlefontsize", "option", "fitcolor",
                "xticks", "yticks", "zticks",
                "xticksformat", "yticksformat", "zticksformat",
                "xtickslabels", "ytickslabels", "ztickslabels",
                "xticksspacing", "yticksspacing", "zticksspacing",
                "xticksfontsize", "yticksfontsize", "zticksfontsize",
                "force_xlim", "force_ylim", "adapt_xlim", "adapt_ylim",
                "xlim", "ylim", "zlim", "legendfontsize", "hide_statbox",
                "hide_fitbox", "hide_clb", "aspect", "figsize", "blank_zero",
                "path", ""]
    tb.check_kwargs(expected, kwargs)
    
    if not self.__up_to_date:
      self._update()
    ax = kwargs.get("ax", None)
    pdf = kwargs.get("pdf", None)
    cm = kwargs.get("cm", plt.cm.plasma)
    xlabel = kwargs.get("xlabel", None)
    ylabel = kwargs.get("ylabel", None)
    zlabel = kwargs.get("zlabel", None)
    xlabelfontsize = kwargs.get("xlabelfontsize", None)
    ylabelfontsize = kwargs.get("ylabelfontsize", None)
    zlabelfontsize = kwargs.get("zlabelfontsize", None)
    title = kwargs.get("title", None)
    titlefontsize = kwargs.get("titlefontsize", None)
    option = kwargs.get("option", "")
    fitcolor = kwargs.get("fitcolor", None)
    xticks = kwargs.get("xticks", None)
    yticks = kwargs.get("yticks", None)
    zticks = kwargs.get("zticks", None)
    xticksformat = kwargs.get("xticksformat", None)
    yticksformat = kwargs.get("yticksformat", None)
    zticksformat = kwargs.get("zticksformat", None)
    xtickslabels = kwargs.get("xtickslabels", None)
    ytickslabels = kwargs.get("ytickslabels", None)
    ztickslabels = kwargs.get("ztickslabels", None)
    xticksspacing = kwargs.get("xticksspacing", None)
    yticksspacing = kwargs.get("yticksspacing", None)
    zticksspacing = kwargs.get("zticksspacing", None)
    xticksfontsize = kwargs.get("xticksfontsize", None)
    yticksfontsize = kwargs.get("yticksfontsize", None)
    zticksfontsize = kwargs.get("zticksfontsize", None)
    force_xlim = kwargs.get("force_xlim", False)
    force_ylim = kwargs.get("force_ylim", False)
    adapt_xlim = kwargs.get("adapt_xlim", False)
    adapt_ylim = kwargs.get("adapt_ylim", False)
    xlim = kwargs.get("xlim", [])
    ylim = kwargs.get("ylim", [])
    zlim = kwargs.get("zlim", [])
    legendfontsize = kwargs.get("legendfontsize", None)
    hide_statbox = kwargs.get("hide_statbox", False)
    hide_fitbox = kwargs.get("hide_fitbox", False)
    hide_clb = kwargs.get("hide_clb", False)
    blank_zero = kwargs.get("blank_zero", False)
    path = kwargs.get("path","")
    aspect = float(
      kwargs.get("aspect", self.get_drange_y()/self.get_drange_x())
    )
    figsize = kwargs.get("figsize", None)
    if pdf is not None and ax is not None:
      raise tb.ArgError("Can not specify a pdf to save to and a subplot"\
                             " to update at the same time.")
    fig = None
    ax_clb = None
    ax_leg = None
    ax1 = ax
    if pdf is not None:
      plt.close("all")
      fig = plt.figure(0)
    if ax1 is None:
      if pdf is None:
        plt.close("all")
      fig, ax1, ax_clb, ax_leg = self.form_subplots_for_clb_and_leg(
        hide_statbox, hide_fitbox, hide_clb, aspect, figsize
      )
    else:
      # Get the current x and y limits if ax specified. If adapt_x/ylim is,
      # specified check if the current limit is too high or too low to display
      # the entire plot and change it if need be.
      xlimlow, xlimhigh = ax1.get_xlim()
      ylimlow, ylimhigh = ax1.get_ylim()
      if Decimal(xlimlow) > self.__arange_x[0] and adapt_xlim:
        ax1.set_xlim(left = float(self.__arange_x[0]))
      if Decimal(xlimhigh) < self.__arange_x[1] and adapt_xlim:
        ax1.set_xlim(right = float(self.__arange_x[1]))
      if Decimal(ylimlow) > self.__arange_y[0] and adapt_ylim:
        ax1.set_ylim(left = float(self.__arange_y[0]))
      if Decimal(ylimhigh) < self.__arange_y[1] and adapt_ylim:
        ax1.set_ylim(right = float(self.__arange_y[1]))
      fig = plt.gcf()
    
    # If x/ylim is specified, set it. Otherwise, if force_xlim is specified,
    # redefine the x/y limites to match the histo ranges.
    if len(xlim) == 2:
      ax1.set_xlim(xlim)
    elif force_xlim:
      ax1.set_xlim([0.99*float(self.__arange_x[0]),
                    1.01*float(self.__arange_x[1])])
    if len(ylim) == 2:
      ax1.set_ylim(ylim)
    elif force_ylim:
      ax1.set_ylim([0.99*float(self.__arange_y[0]),
                    1.01*float(self.__arange_y[1])])
    
    vmin = None
    vmax = None
    if blank_zero:
      vmin = 1e-10
      cm.set_under(color="white")
      if np.amax(self.__bin_contents) < vmin:
        vmax = 1
    im = ax1.imshow(
      np.rot90(self.__bin_contents),
      interpolation="nearest",
      cmap=cm,
      extent=(float(self.__arange_x[0]), float(self.__arange_x[1]),
              float(self.__arange_y[0]), float(self.__arange_y[1])),
      aspect=aspect*(float(self.__arange_x[1]-self.__arange_x[0])
                    /float(self.__arange_y[1]-self.__arange_y[0])),
      vmin=vmin,
      vmax=vmax
    )
    if len(zlim) == 2:
      im.set_clim(zlim[0], zlim[1])
      
    if xlabel is not None:
      ax1.set_ylabel(ylabel, fontsize=ylabelfontsize)
    if ylabel is not None:
      ax1.set_xlabel(xlabel, fontsize=xlabelfontsize)
    if title is not None:
      ax1.set_title(title, fontsize = titlefontsize)
    clb = None
    if ax_clb is not None:
      if aspect <= 1.5:
        clb = plt.colorbar(im, cax=ax_clb, cmap=cm, ticks=zticks,
                           format=zticksformat)
      else:
        clb = plt.colorbar(im, cmap=cm, cax=ax_clb, ticks=zticks,
                           format=zticksformat, orientation = "horizontal")
      if zlabel is not None:
        clb.set_label(zlabel, fontsize = zlabelfontsize)
    elif ax is not None and not hide_clb:  # ax1 exists now, so use ax to test
                                          # if arg was given.
      if aspect <= 1.5:
        clb = plt.colorbar(im, ax=ax1, cmap=cm, ticks=zticks)
      else:
        clb = plt.colorbar(im, cmap=cm, ax=ax1, ticks=zticks,
                           format=zticksformat, orientation = "horizontal")
      if zlabel is not None:
        clb.set_label(zlabel, fontsize = zlabelfontsize)
    fig.canvas.draw()
    renderer = fig.canvas.renderer
    if ax_leg is not None:
      props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
      leg = ax_leg.text(0.15, 0.95, self.__statbox, transform=ax_leg.transAxes,
                  fontsize=legendfontsize, verticalalignment='top',
                  bbox=props)
      # Adjust the legend box to fit the subplot
      # Returns [[x0,y0],[x1,y1]]
      ax_xy0, ax_xy1 = ax_leg.get_window_extent(renderer).get_points()
      leg_xy0, leg_xy1 = leg.get_window_extent(renderer).get_points()
      att = 0
      while (ax_xy0[0] > leg_xy0[0] or ax_xy0[1] > leg_xy0[1]
          or ax_xy1[0] < leg_xy1[0] or ax_xy1[1] < leg_xy1[1]) and att < 100:
        att += 1
        leg._fontproperties._size = leg._fontproperties._size*0.85
        leg_xy0, leg_xy1 = leg.get_window_extent(renderer).get_points()
      if att == 100:
        warnings.warn("Legend might be too big", Warning)
    
    if xticksfontsize is not None:
      ax1.tick_params(axis='x', labelsize=xticksfontsize)
    if yticksfontsize is not None:
      ax1.tick_params(axis='y', labelsize=yticksfontsize)
    if zticksfontsize is not None:
      clb.ax.yaxis.set_tick_params(labelsize=zticksfontsize)
    plt.gcf().subplots_adjust(left=0.2)
      
    if xticks is not None:
      if xticks == 'bins':
        xticks = self.__bin_edges_x
      ax1.set_xticks(xticks)
    if yticks is not None:
      if yticks == 'bins':
        yticks = self.__bin_edges_y
      ax1.set_yticks(yticks)
    if xticksspacing is not None:
      ax1.xaxis.set_major_locator(ticker.MultipleLocator(xticksspacing))
    if yticksspacing is not None:
      ax1.yaxis.set_major_locator(ticker.MultipleLocator(yticksspacing))
    if xticksformat is not None:
      ax1.xaxis.set_major_formatter(ticker.FormatStrFormatter(xticksformat))
    if yticksformat is not None:
      ax1.yaxis.set_major_formatter(ticker.FormatStrFormatter(yticksformat))
    if xtickslabels is not None:
      ax1.set_xticklabels(xtickslabels)
    if ytickslabels is not None:
      ax1.set_yticklabels(ytickslabels)
#    plt.gcf().tight_layout()
    
    if "save" in option and pdf is None:
      path = os.path.join(
          path,
          str("").join(("MyHist2D_", self.__name, ".pdf"))
        )
      plt.savefig(path, bbox_inches="tight")
    if "show" in option:
      plt.show()
    if pdf is not None:
      pdf.savefig(fig, bbox_inches="tight")
      plt.close('all')
    else:
      return fig
